from poreana.sample import Sample

from poreana.model import Model
from poreana.model import CosineModel
from poreana.model import StepModel
from poreana.model import StepModel

from poreana.mc import MC

import poreana.adsorption as adsorption
import poreana.density as density
import poreana.diffusion as diffusion
import poreana.freeenergy as freeenergy
import poreana.tables as tables
import poreana.gyration as gyration

import poreana.geometry as geom
import poreana.utils as utils


__all__ = [
    "Sample",
    "Model", "CosineModel", "StepModel", "MC",
    "adsorption", "density", "diffusion", "freeenergy",
    "geom", "utils","tables"
]



################################################################################
# Utils                                                                        #
#                                                                              #
"""Here popular basic methods are noted."""
################################################################################


import os
import time
import h5py
import yaml
import pickle
import datetime
import numpy as np
import pandas as pd
import poreana as pa


def mkdirp(directory):
    """Create directory if it does not exist.

    Parameters
    ----------
    directory : string
        Directory name
    """
    if not os.path.exists(directory):
        os.makedirs(directory)


def column(data):
    """Convert given row list matrix into column list matrix

    Parameters
    ----------
    data : list
        Row data matrix

    Returns
    -------
    data_col : list
        column data matrix
    """
    num_row = len(data)
    num_col = len(data[0])

    data_col = [[] for i in range(num_col)]

    for i in range(num_row):
        for j in range(num_col):
            data_col[j].append(data[i][j])

    return data_col


def tic():
    """MATLAB tic replica - return current time.

    Returns
    -------
    time : float
        Current time in seconds
    """
    return time.time()


def toc(t, message="", is_print=True):
    """MATLAB toc replica - return time difference to tic and alternatively
    print a message.

    Parameters
    ----------
    t : float
        Starting time - given by tic function
    message : string, optional
        Custom output message
    is_print : bool, optional
        True for printing an output message

    Returns
    -------
    time : float
        Time difference
    """
    if message:
        message += " - runtime = "

    t_diff = time.time()-t

    if is_print:
        print(message+"%6.3f" % t_diff+" s")

    return t_diff


def save(obj, link):
    """Save an object files, yaml files, and hd5 files using pickle in the specified link.

    Parameters
    ----------
    obj : Object
        Object to be saved
    link : string
        Specific link to save object
    """
    # Pickle object file


    # Hd5 file
    if link.split(".")[-1]=="h5":
        # Save results in a hdf5 file
        f = h5py.File(link, 'w')

        # Create input groupe
        groups = {}
        data_base = {}

        # Create groups for the keys in the obj file
        for key in obj.keys():
            groups[key] = f.create_group(key)

            # If key "box" only a value is on frist key level
            if not key in ["box","type"]:
            # For all other key read the second level keys
                data_base[key] = obj[key].keys()

        for gkey in groups:
            # Write the box length on data base
            if gkey =="box":
                groups[gkey].create_dataset("length",data=obj[gkey]["length"])
            # Write the type of calculation on data base
            elif gkey =="type":
                dt = h5py.special_dtype(vlen=str)
                type = groups[gkey].create_dataset("type", (1), dtype=dt)
                type[0] = obj[gkey]
            # Loop over second level keys
            else:
                for base in data_base[gkey]:
                    # Check if a third level exists
                    if isinstance(obj[gkey][base], dict):
                        # If a third level exists create new group in first level groups
                        data = groups[gkey].create_group(base)
                        # Loop over thrid level keys
                        for base2 in obj[gkey][base]:
                            # Check if third level key is a matrix
                            if isinstance(obj[gkey][base][base2], (list, np.ndarray)):
                                data.create_dataset(str(base2), data=obj[gkey][base][base2])
                            # Else is a vlaue (int/float)
                            else:
                                value = data.create_dataset(str(base2), shape=(1,1))
                                value[0] = obj[gkey][base][base2]
                    # Check If second level is value/list or string
                    # If string
                    elif isinstance(obj[gkey][base],str):
                        dt = h5py.special_dtype(vlen=str)
                        string = groups[gkey].create_dataset(str(base), (1), dtype=dt)
                        string[0] = obj[gkey][base]
                    # If list
                    elif isinstance(obj[gkey][base], (list, np.ndarray)):
                        data = groups[gkey].create_dataset(str(base), data=obj[gkey][base])
                    # If value
                    else:
                        data = groups[gkey].create_dataset(str(base), shape=(1,1))
                        data[0] = obj[gkey][base]

    elif link.split(".")[-1]=="yml":
        with open(link, "w") as file_out:
            file_out.write(yaml.dump(obj))
    else:
        if link.split(".")[-1]!="obj":
            print("Wrong data type")
            return

        with open(link, "wb") as f:
            pickle.dump(obj, f)

def load(link, file_type=""):
    """Load pickled object files, yaml files, and hd5 files.

    Parameters
    ----------
    link : string
        Specific link to load object
    file_type : string, optional
        Specify filetype - **obj**, **yml**, **h5**  leave empty for automatic
        determination

    Returns
    -------
    obj : Object
        Loaded object
    """
    # YAML file
    if file_type=="yml" or (not file_type and link.split(".")[-1]=="yml"):
        with open(link, "r") as file_in:
            return yaml.load(file_in, Loader=yaml.UnsafeLoader)

    # Hd5 file
    elif file_type=="h5" or (not file_type and link.split(".")[-1]=="h5"):
        data = h5py.File(link, 'r')
        data_load = {}

        # Check keys in hdf5 file
        for keys in data.keys():
            data_load[keys] = {}
            # If sample file is load
            if keys=="data":
                for keys2 in data[keys].keys():
                    try:
                        data_load[keys][int(keys2)] = data[keys][keys2][:]
                    except(ValueError):
                        data_load[keys][keys2] = data[keys][keys2][:]
            # If key type only a string has to be load
            elif keys=="type":
                data_load[keys] = str(data[keys][keys][0], 'utf-8') #data[keys][keys][0].encode().decode("utf-8")
            # other keys
            else:
                # Load second level keys
                for keys2 in data[keys].keys():
                    data_load[keys][keys2] = {}
                    # Try if a thrid level exist
                    try:
                        # load thrid level
                        for keys3 in data[keys][keys2].keys():
                            data_load[keys][keys2][int(keys3)] = data[keys][keys2][keys3][:]
                    # Save second level data
                    except(AttributeError):
                        if len(data[keys][keys2][:])==1:
                            try:
                                data_load[keys][keys2] = float(data[keys][keys2][0])
                            except(ValueError):
                                data_load[keys][keys2] = str(data[keys][keys2][0], 'utf-8') #.encode().decode("utf-8")
                        else:
                            data_load[keys][keys2] = data[keys][keys2][:]
        return data_load
    else:
        if link.split(".")[-1]!="obj":
            print("Wrong data type")
            return
        with open(link, 'rb') as f:
            return pickle.load(f)



def file_to_text(link, link_output, link_dens=[]):
    """This function converts an output directory in txt file. For the bin diffusion
    result text file the density sampling file is additionally necessary.

    Parameters
    ----------
    link : string
        Link to output hdf5 or obj file
    link_output : String
        Link to an output text file
    link_dens : optional
        Link to density output hdf5 or obj file. Only necessary for the bin
        diffusion and gyration text ouput.
    """

    # Load data
    data = load(link)

    ###############################
    # Further properties Function #
    ###############################
    if data["type"] == "gyr_bin":
        if not link_dens:
            print("Gyration calculation needs a density sampling file. Check documentation and set a link_dens.")
            return

        # If pore system
        if "pore" in data:
            # Read data
            pore = data["pore"]
            res = float(pore["res"])
            diam = float(pore["diam"])
            box = pore["box"]
            type = pore["type"]

            # Set panda tables for pore properties
            data_system = [[["%.2f" % i for i in box]],[diam],[res],[type]]
            df_system = pd.DataFrame(data_system, index=list(["Box dimension (nm)","Pore diameter (nm)","reservoir (nm)", "type"]),columns=list(["Value"]))
            df_system = df_system.rename_axis('# Identifier', axis=1)
        # If box system
        elif "box" in data:
            box_group = data["box"]
            box = box_group["length"]
            data_box = [[["%.2f" % i for i in box]]]
            df_system = pd.DataFrame(data_box, index=list(["Box dimension (nm)"]), columns=list(["Value"]))
            df_system = df_system.rename_axis('# Identifier', axis=1)

        # Set pandas table for input
        data = [[data["inp"]["bin_num"]],[data["inp"]["entry"]],[data["inp"]["num_frame"]], [data["inp"]["mass"]]]
        df_inputs = pd.DataFrame(data,index = list(["Bin number","Entry","Frame number", "Mass"]), columns=list(["Value"]))
        df_inputs = df_inputs.rename_axis('# Identifier', axis=1)

        # Calculate gyration and set pandas table
        gyr_in = pa.gyration.bins_plot(link, link_dens, intent="ex")
        gyr_pd = pd.DataFrame(gyr_in, index=(["gyration"]))
        gyr_pd = gyr_pd.rename_axis('# Identifier', axis=1)

        # Convert pandas table to string
        df_inputs_string = df_inputs.to_string(header=True, index=True)
        df_system_string = df_system.to_string(header=True, index=True)
        gyr_pd_string = gyr_pd.to_string(header=True, index=True)

        # Write file
        with open(link_output, 'w') as file:
            # Write file header
            file.write("# This file was created " + str(datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")) + "\n")
            file.write("# Analysis created using  PoreAna\n")
            file.write("# Object file : " + os.path.dirname(os.path.abspath(__file__)) + link + "\n\n")

            # Write System table
            file.write("[System]\n")
            file.write(df_system_string)

            # Write Inputs
            file.write("[Input]\n")
            file.write(df_inputs_string)

            # Write gyration table
            file.write("\n\n[Gyration]\n")
            file.write(gyr_pd_string)
        # Close file
        file.close()


    ##########################
    # Bin Diffusion Function #
    ##########################
    # If diffusion bin sample file was loaded
    elif data["type"] == "diff_bin":
        # Check inputs (for bin diffusion is density sample file necessary)
        if not link_dens:
            print("Bin diffusion needs a density sampling file. Check documentation and set a link_dens.")
            return

        # Calculate bin diffusion
        bins = pa.diffusion.bins(link, is_norm=True)
        dens = pa.density.bins(link_dens)
        mean = pa.diffusion.mean(bins, dens)

        # If pore system
        if "pore" in data:
            # Read data
            pore = data["pore"]
            res = float(pore["res"])
            diam = float(pore["diam"])
            box = pore["box"]
            type = pore["type"]

            # Set panda tables for pore properties
            data_system = [[["%.2f" % i for i in box]],[diam],[res],[type]]
            df_system = pd.DataFrame(data_system, index=list(["Box dimension (nm)","Pore diameter (nm)","reservoir (nm)", "type"]),columns=list(["Value"]))
            df_system = df_system.rename_axis('# Identifier', axis=1)


        # Set pandas table for input
        data_inp = [[data["inp"]["bin_num"]],[data["inp"]["entry"]],[data["inp"]["num_frame"]], [data["inp"]["mass"]], [bins["is_norm"]]]
        df_inputs = pd.DataFrame(data_inp,index = list(["Bin number","Entry","Frame number", "Mass", "is_norm"]), columns=list(["Value"]))
        df_inputs = df_inputs.rename_axis('# Identifier', axis=1)

        # Set pandas table for bin diffusion profiles
        data = {}
        data["# Width"] = bins["width"][1:]
        data["D"] = bins["diff"]
        df_data = pd.DataFrame(data)

        # Pandas table to string
        df_system_string = df_system.to_string(header=True, index=True)
        df_inputs_string = df_inputs.to_string(header=True, index=True)
        df_data_string = df_data.to_string(header=True, index=False)

        # Write file
        with open(link_output, 'w') as file:
            # Write file header
            file.write("# This file was created " + str(datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")) + "\n")
            file.write("# Analysis created using  PoreAna\n")
            file.write("# Object file : " + os.path.dirname(os.path.abspath(__file__)) + link + "\n")
            file.write("# Units\n")
            file.write("# Diffusion D(10^-9 m^2s^-1)\n\n")

            # Write System table
            file.write("[System]\n")
            file.write(df_system_string)

            # Wirte input table
            file.write("\n\n[Inputs]\n")
            file.write(df_inputs_string)

            # Write diffusion coefficent
            file.write("\n\n[Diffusion]\n")
            file.write(str("%.2f " % mean ) + "* 10^-9 m^2s^-1")

            # Write diffusion profile
            file.write("\n\n[Bin Diffusion Profiles]\n")
            file.write(df_data_string)

        # Close file
        file.close()

    ####################
    # Density Function #
    ####################
    # If density sample file was loaded
    elif data["type"] == "dens_bin":
        # If pore system
        if "pore" in data:
            pore = data["pore"]
            res = float(pore["res"])
            diam = float(pore["diam"])
            box = pore["box"]
            type = pore["type"]

            # Set panda table for pore properties
            data_system = [[["%.2f" % i for i in box]],[diam],[res],[type]]
            df_system = pd.DataFrame(data_system, index=list(["Box dimension (nm)","Pore diameter (nm)","reservoir (nm)", "type"]),columns=list(["Value"]))
            df_system = df_system.rename_axis('# Identifier', axis=1)
        # If box system
        elif "box" in data:
            box_group = data["box"]
            box = box_group["length"]
            data_box = [[["%.2f" % i for i in box]]]
            df_system = pd.DataFrame(data_box, index=list(["Box dimension (nm)"]), columns=list(["Value"]))
            df_system = df_system.rename_axis('# Identifier', axis=1)

        # Set panda table for inputs
        data_inp = [[data["inp"]["bin_num"]],[data["inp"]["entry"]],[data["inp"]["num_frame"]],[True if data["inp"]['remove_pore_from_res'] else False], [str("%.2f" % data["inp"]["mass"])]]
        df_inputs = pd.DataFrame(data_inp,index = list(["Bin number","Entry","Frame number", "Remove pore from reservoir", "Mass"]), columns=list(["Value"]))
        df_inputs = df_inputs.rename_axis('# Identifier', axis=1)

        # Calculated adsorption and set pandas table
        if "pore" in data:
            ads = pa.adsorption.calculate(link)
            df_ads = pd.DataFrame(ads)
            df_ads = df_ads.rename_axis('# Identifier', axis=1)

        # Calculated density and set pandas table
        dens = pa.density.bins(link)
        if  "pore" in data:
            df_mean = pd.DataFrame([[dens["mean"]["in"], dens["dens"]["in"]],[dens["mean"]["ex"], dens["dens"]["ex"]]],index = list(["Density inside pore","Density outside pore"]), columns=list(["Density (#/nm^3)", "Density (kg/m^3)"]))
        else:
            df_mean = pd.DataFrame([[dens["mean"]["ex"], dens["dens"]["ex"]]],index = list(["Density box"]), columns=list(["Density (#/nm^3)", "Density (kg/m^3)"]))
        df_mean = df_mean.rename_axis('# Identifier', axis=1)

        # Set pandas table for density profile
        data_dens = {}
        data_dens["# Ex width"] = dens["sample"]["data"]["ex_width"]
        data_dens["Density (Ex)"] = dens["num_dens"]["ex"]
        if "pore" in data:
            data_dens["In width"] = dens["sample"]["data"]["in_width"][1:]
            data_dens["Density (In)"] = dens["num_dens"]["in"]
        df_data = pd.DataFrame(data_dens)

        # Convert pandas table to string
        if "pore" in data:
            df_ads_string = df_ads.to_string(header=True, index=True, na_rep="")
        df_data_string = df_data.to_string(header=True, index=False)
        df_mean_string = df_mean.to_string(header=True, index=True)
        df_system_string = df_system.to_string(header=True, index=True)
        df_inputs_string = df_inputs.to_string(header=True, index=True)

        # Write file
        with open(link_output, 'w') as file:
            # Write tables in file
            file.write("# This file was created " + str(datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")) + "\n")
            file.write("# Analysis created using  PoreAna\n")
            file.write("# Object file : " + os.path.dirname(os.path.abspath(__file__)) + link + "\n\n")

            # Write system table
            file.write("[System]\n")
            file.write(df_system_string)

            # Write input table
            file.write("\n\n[Inputs]\n")
            file.write(df_inputs_string)

            # Write adsorption table
            if "pore" in data:
                file.write("\n\n[Adsorption]\n")
                file.write(df_ads_string)

            # Write density
            file.write("\n\n[Density]\n")
            file.write(df_mean_string)

            # Write density profile
            file.write("\n\n[Density Profiles]\n")
            file.write(df_data_string)

        # Close file
        file.close()

    ###############
    # MC Function #
    ###############
    # If mc file was loaded
    elif data["type"] == "mc":
        # If pore system
        if "pore" in data:
            pore = data["pore"]
            t = data["model"]["len_frame"]
            res = float(pore["res"])
            diam = float(pore["diam"])
            box = pore["box"]
            type = pore["type"]
            data = [[["%.2f" % i for i in box]],[diam],[res],[type]]
            df_system = pd.DataFrame(data, index=list(["Box dimension (nm)","Pore diameter (nm)","reservoir (nm)", "type"]),columns=list(["Value"]))
            df_system = df_system.rename_axis('# Identifier', axis=1)

        # If box system
        elif "box" in data:
            t = data["model"]["len_frame"]
            box_group = data["box"]
            box = box_group["length"]
            data = [[["%.2f" % i for i in box]]]
            df_system = pd.DataFrame(data, index=list(["Box dimension (nm)"]), columns=list(["Value"]))
            df_system = df_system.rename_axis('# Identifier', axis=1)


        # Get profiles
        diff = pa.diffusion.mc_profile(link, is_plot = False, infty_profile=True)
        free_energy= pa.freeenergy.mc_profile(link, is_plot = False)

        # Set pandas table for profiles
        data = {}
        data["# Bins [nm]"] = diff[2]
        for i in diff[1]:
            data["D (t={})".format(t*i)] = diff[1][i]
        data["   D (t=∞)"]=diff[0]
        for i in free_energy[0].keys():
            data["Free energy [-]"]= free_energy[0][i][1:]
        df_data = pd.DataFrame(data)


        # Set pandas tables
        df_inputs = pa.tables.mc_inputs(link, print_con=False)
        df_inputs = df_inputs.rename_axis('# Identifier', axis=1)
        df_model = pa.tables.mc_model(link, print_con=False)
        df_model = df_model.rename_axis('# Identifier', axis=1)
        df_results = pa.tables.mc_results(link, print_con=False)
        df_results = df_results.rename_axis('# Identifier', axis=1)

        # Convert pandas table to strings
        df_model_string = df_model.to_string(header=True, index=True)
        df_inputs_string = df_inputs.to_string(header=True, index=True)
        df_results_string = df_results.to_string(header=True, index=True)
        df_system_string = df_system.to_string(header=True, index=True)
        df_data_string = df_data.to_string(header=True, index=False)

        # Write output file
        with open(link_output, 'w') as file:
            # Write tables in file
            file.write("# This file was created " + str(datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")) + "\n")
            file.write("# Analysis created using  PoreAna\n")
            file.write("# Object file : " + os.path.dirname(os.path.abspath(__file__)) + "/" + link + "\n")
            file.write("# Units\n")
            file.write("# Diffusion D(10^-9 m^2s^-1)\n")
            file.write("# Lag time  t(ps)\n\n")

            # Write system table
            file.write("[System]\n")
            file.write(df_system_string)

            # Write model input table
            file.write("\n\n\n[Model Inputs]\n")
            file.write(df_model_string)

            # Write mc input table
            file.write("\n\n\n[MC Inputs]\n")
            file.write(df_inputs_string)

            # Write MC results (Diffusion coefficients)
            file.write("\n\n\n[MC Results]\n")
            file.write(df_results_string)

            # Write diffusion and free energy profile in file
            file.write("\n\n[Profiles]\n")
            file.write(df_data_string)

            # Close file
            file.close()

def mumol_m2_to_mols(c, A):
    """Convert the concentration in :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`
    to number of molecules.

    The concentration is given as

    .. math::

        c=\\frac{N}{N_A\\cdot A}

    with surface :math:`A` and avogadro constant
    :math:`N_A=6.022\\cdot10^{23}\\,\\text{mol}^{-1}`. Assuming that the unit of
    the concentration is :math:`\\mu\\text{mol}\\cdot\\text{m}^{-2}` and of the
    surface is :math:`\\text{nm}^2`, the number of molecules is given by

    .. math::

        N=c\\cdot N_A\\cdot A
        =[c]\\cdot10^{-24}\\,\\frac{\\text{mol}}{\\text{nm}^2}\\cdot6.022\\cdot10^{23}\\,\\frac{1}{\\text{mol}}\\cdot[A]\\,\\text{nm}^2

    and thus

    .. math::

        \\boxed{N=0.6022\\cdot[c]\\cdot[A]}\\ .

    Parameters
    ----------
    c : float
        Concentration in :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`
    A : float
        Surface in :math:`\\text{nm}^2`

    Returns
    -------
    N : float
        Number of molecules
    """
    return 0.6022*c*A


def mols_to_mumol_m2(N, A):
    """Convert the number of molecules to concentration in
    :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`.

    The concentration is given as

    .. math::

        c=\\frac{N}{N_A\\cdot A}

    with surface :math:`A` and avogadro constant
    :math:`N_A=6.022\\cdot10^{23}\\,\\text{mol}^{-1}`. Assuming that the unit of
    the concentration is :math:`\\mu\\text{mol}\\cdot\\text{m}^{-2}` and of the
    surface is :math:`\\text{nm}^2`, the number of molecules is given by

    .. math::

        c=\\frac{[N]}{6.022\\cdot10^{23}\\,\\text{mol}^{-1}\\cdot[A]\\,\\text{nm}^2}

    and thus

    .. math::

        \\boxed{c=\\frac{N}{0.6022\\cdot[A]}\\cdot\\frac{\\mu\\text{mol}}{\\text{m}^2}}\\ .

    Parameters
    ----------
    N : int
        Number of molecules
    A : float
        Surface in :math:`\\text{nm}^2`

    Returns
    -------
    c : float
        Concentration in :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`
    """
    return N/0.6022/A


def mmol_g_to_mumol_m2(c, SBET):
    """Convert the concentration :math:`\\frac{\\text{mmol}}{\\text{g}}`
    to :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`.

    This is done by dividing the concentration per gram by the material surface
    per gram property :math:`S_\\text{BET}`

    .. math::

        \\boxed{c_A=\\frac{c_g}{S_\\text{BET}}\\cdot10^3}\\ .

    Parameters
    ----------
    c : float
        Concentration in :math:`\\frac{\\text{mmol}}{\\text{g}}`
    SBET : float
        Material surface in :math:`\\frac{\\text{m}^2}{\\text{g}}`

    Returns
    -------
    c : float
        Concentration in :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`
    """
    return c/SBET*1e3


def mmol_l_to_mols(c, V):
    """Convert the concentration in :math:`\\frac{\\text{mmol}}{\\text{l}}`
    to number of molecules.

    The concentration in regard to volume is calculated by

    .. math::

        c=\\frac{N}{N_A\\cdot V}

    with volume :math:`V`. Assuming that the unit of the concentration is
    :math:`\\text{mmol}\\cdot\\text{l}` and of the volume is
    :math:`\\text{nm}^3`, the number of molecules is given by

    .. math::

        N=c\\cdot N_A\\cdot V
        =[c]\\cdot10^{-27}\\,\\frac{\\text{mol}}{\\text{nm}^3}\\cdot6.022\\cdot10^{23}\\,\\frac{1}{\\text{mol}}\\cdot[V]\\,\\text{nm}^3

    and thus

    .. math::

        \\boxed{N=6.022\\cdot10^{-4}\\cdot[c]\\cdot[V]}\\ .

    Parameters
    ----------
    c : float
        Concentration in :math:`\\frac{\\text{mmol}}{\\text{l}}`
    V : float
        Surface in :math:`\\text{nm}^3`

    Returns
    -------
    N : float
        Number of molecules
    """
    return 6.022e-4*c*V


def mols_to_mmol_l(N, V):
    """Convert the number of molecules to concentration in
    :math:`\\frac{\\text{mmol}}{\\text{l}}`.

    The concentration in regard to volume is calculated by

    .. math::

        c=\\frac{N}{N_A\\cdot V}

    with volume :math:`V`. Assuming that the unit of the concentration is
    :math:`\\text{mmol}\\cdot\\text{l}` and of the volume is
    :math:`\\text{nm}^3`, the number of molecules is given by

    .. math::

        c=\\frac{N}{6.022\\cdot10^{23}\\cdot[V]\\,\\text{nm}^3}

    and thus

    .. math::

        \\boxed{c=\\frac{N}{6.022\\times10^{-4}\\cdot[V]}\\cdot\\frac{\\text{mmol}}{\\text{l}}}\\ .

    Parameters
    ----------
    N : int
        Number of molecules
    V : float
        Surface in :math:`\\text{nm}^3`

    Returns
    -------
    c : float
        Concentration in :math:`\\frac{\\text{mmol}}{\\text{l}}`
    """
    return N/6.022e-4/V



################################################################################
# Diffusion                                                                    #
#                                                                              #
"""Analyse diffusion in a pore."""
################################################################################


import math
import itertools
import scipy as sp
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

import poreana.utils as utils


###################
# Diffusion - Bin #
###################
def cui(link, z_dist=0, ax_area=[0.2, 0.8], intent="", is_fit=False, is_plot=True):
    """This function samples and calculates the diffusion coefficient of a
    molecule group in a pore in both axial and radial direction, as described
    in the paper of `Cui <https://doi.org/10.1063/1.1989314>`_.

    The mean square displacement is sampled in function
    :func:`poreana.sample.Sample._diffusion_bin`.

    The axial diffusion is given by the Einstein relation

    .. math::

        \\langle\\left[z(0)-z(t)\\right]^2\\rangle=2D_\\text{axial}t

    with axial diffusion coefficient :math:`D_\\text{axial}`. Thus the
    coefficient corresponds to the slope of the axial msd

    .. math::

        D_\\text{axial}=\\frac12\\frac{\\text{msd}_\\text{axial}[i]-\\text{msd}_\\text{axial}[j]}{t_i-t_j}

    with bin :math:`i>j`. The radial diffusion is given by

    .. math::

        \\langle\\left[r(0)-r(t)\\right]^2\\rangle=R^2\\left[1-\\sum_{n=1}^\\infty\\frac{8}{\\lambda_{1n}^2(\\lambda_{1n}^2-1)}\\exp\\left(-\\frac{\\lambda_{1n}^2}{R^2}D_\\text{radial}t\\right)\\right].

    with radial diffusion coefficient :math:`D_\\text{radial}`, the maximal
    accessible radial position :math:`R` by an atom

    .. math::

        R = \\frac12d-0.2

    with pore diameter :math:`d`, and the zeros :math:`\\lambda_{1n}^2` of
    the derivative

    .. math::

        \\frac{dJ_1}{dx}

    of the first order Bessel function :math:`J_1`. Therefore the coefficient
    has to be fitted to the sampled radial msd. The author observed that
    the first 20 Bessel function zeros were sufficient for the function fit.

    The final unit transformation is done by

    .. math::

        \\frac{\\text{nm}^2}{\\text{ps}}=10^{-2}\\frac{\\text{cm}^2}{\\text{s}}.

    **Note that the function for the radial diffusion is obtained under the
    assumption that the density inside the pore is uniform.**

    Parameters
    ----------
    link : string
        Link to hdf5 data file generated by the sample routine
        :func:`poreana.sample.Sample._diffusion_bin`
    z_dist : float, optional
        Distance from pore centre to calculate the mean
    ax_area : list, optional
        Bin area percentage to calculate the axial diffusion coefficient
    intent : string, optional
        axial, radial or empty for both
    is_fit : bool, optional
        True to plot the fitted function
    is_plot : bool, optional
        True to create plot in this function
    """
    # Load data object
    sample = utils.load(link)

    # Load data
    pore = sample["pore"]
    inp = sample["inp"]
    len_window = int(inp["len_window"])
    len_step = inp["len_step"]
    len_frame = inp["len_frame"]
    bins = int(inp["bin_num"]) if not z_dist else math.floor(z_dist/sample["data"]["width"])
    msd_z = [0 for x in range(len_window)]
    msd_r = [0 for x in range(len_window)]
    norm_z = [0 for x in range(len_window)]
    norm_r = [0 for x in range(len_window)]

    # Sum up all bins
    for i in range(bins):
        for j in range(len_window):
            msd_z[j] += sample["data"]["z_tot"][i][j]
            norm_z[j] += sample["data"]["n_tot"][i][j]

    for i in range(int(inp["bin_num"])):
        for j in range(len_window):
            msd_r[j] += sample["data"]["r_tot"][i][j]
            norm_r[j] += sample["data"]["n_tot"][i][j]

    # Normalize
    msd_z_n = [msd_z[i]/norm_z[i] if norm_z[i] > 0 else 0 for i in range(len_window)]
    msd_r_n = [msd_r[i]/norm_r[i] if norm_r[i] > 0 else 0 for i in range(len_window)]

    # Define time axis and range
    time_ax = [x*len_step*len_frame for x in range(len_window)]
    t_range = (len_window-1)*len_step*len_frame

    # Calculate axial coefficient
    if not intent or intent == "axial":
        dz = (msd_z_n[int(ax_area[1]*len_window)]-msd_z_n[int(ax_area[0]*len_window)])*1e-9**2/((ax_area[1]-ax_area[0])*t_range)/2*1e2**2*1e5  # 10^-9 m^2s^-1

        print("Diffusion axial:  "+"%.3f" % dz+" 10^-9 m^2s^-1")

    # Calculate radial coefficient
    if not intent or intent == "radial":
        def diff_rad(x, a, b, c):
            # Process input
            x = x if isinstance(x, list) or isinstance(x, np.ndarray) else [x]

            # Get bessel function zeros
            jz = sp.special.jnp_zeros(1, math.ceil(b))
            # Calculate sum
            sm = [[8/(z**2*(z**2-1))*math.exp(-(z/c)**2*a*t) for z in jz] for t in x]
            # Final equation
            return [c**2*(1-sum(s)) for s in sm]

        # Fit function
        popt, pcov = sp.optimize.curve_fit(diff_rad, [x*1e12 for x in time_ax], msd_r_n, p0=[1, 20, float(pore["diam"])/2-0.2], bounds=(0, np.inf))

        print("Diffusion radial: "+"%.3f" % (popt[0]*1e3)+" 10^-9 m^2 s^-1; Number of zeros: "+"%2i" % (math.ceil(popt[1]))+"; Radius: "+"%5.2f" % popt[2])

    # Plot
    if is_plot:
        legend = []

    if not intent or intent == "axial":
        sns.lineplot(x=[x*1e12 for x in time_ax], y=msd_z_n)
        if is_plot:
            legend += ["Axial"]
        if is_fit:
            sns.lineplot(x=[x*1e12 for x in time_ax], y=[dz*2*time_ax[x]/1e5/1e-7**2 for x in range(len_window)])
            legend += ["Fitted Axial"]

    if not intent or intent == "radial":
        sns.lineplot(x=[x*1e12 for x in time_ax], y=msd_r_n)
        if is_plot:
            legend += ["Radial"]
        if is_fit:
            sns.lineplot(x=[x*1e12 for x in time_ax], y=diff_rad([x*1e12 for x in time_ax], *popt))
            legend += ["Fitted Radial"]

    if is_plot:
        plt.xlabel("Time (ps)")
        plt.ylabel(r"Mean square displacement (nm$^2$)")
        plt.legend(legend)


def bins(link, ax_area=[0.2, 0.8], is_norm=False):
    """This function calculates the axial (z-axis) diffusion coefficient as a
    function of the radial distance. This is done by sampling the mean square
    displacement for all molecules in a radial sub volume.

    The mean square displacement is sampled in function
    :func:`poreana.sample.Sample._diffusion_bin`.

    For each bin, the msd is summed up, resulting into a msd slope for each
    bin. Thus, the axial diffusion coefficient can be calculated using

    .. math::

        D_\\text{axial}=\\frac12\\frac{\\text{msd}_\\text{axial}[i]-\\text{msd}_\\text{axial}[j]}{t_i-t_j}.

    Note that the msd is evaluated in the area, where the slope is uniform,
    which means that the first twenty and last twenty percent should be
    neglected.

    If ``is_norm`` is set to **True**, the radius will be normalized in respect
    to the effective radius which means, the last radius that has a
    Diffusion greater than zero is taken

    .. math::

        r_\\text{norm}=\\frac{1}{r_\\text{eff}}r.

    Parameters
    ----------
    link : string
        Link to hdf5 data file generated by the sample routine :func:`poreana.sample.Sample._diffusion_bin`
    ax_area : list, optional
        Bin area percentage to calculate the axial diffusion coefficient
    intent : string, optional
        Set to **plot**, for plotting or set to **line** to only return the
        lineplot, leave empty for nothing
    is_norm : bool, optional
        True to normalize x-axis

    Returns
    -------
    diffusion : list
        List of the slope of the non-normalized diffusion coefficient
    """
    # Load data object
    sample = utils.load(link)

    # Load data
    inp = sample["inp"]
    width = sample["data"]["width"]
    msd_z = sample["data"]["z"]
    norm = sample["data"]["n"]
    len_window = int(inp["len_window"])
    len_step = inp["len_step"]
    len_frame = inp["len_frame"]


    # Normalize
    msd_norm = [[msd_z[i][j]/norm[i][j] if norm[i][j] > 0 else 0 for j in range(len_window)] for i in range(int(inp["bin_num"])+1)]

    # Calculate slope
    f_start = int(ax_area[0]*len_window)
    f_end = int(ax_area[1]*len_window)
    time_ax = [x*len_step*len_frame for x in range(len_window)]
    slope = [(msd_norm[i][f_end]-msd_norm[i][f_start])/(time_ax[f_end]-time_ax[f_start]) for i in range(int(inp["bin_num"])+1)]

    # Calculate diffusion coefficient
    diff = [msd*1e-9**2/2*1e2**2*1e5 for msd in slope]  # 10^-9 m^2s^-1

    return {"width": width, "diff": diff, "is_norm": is_norm}


def bins_plot(data, intent="plot", kwargs={}):
    """This function calculates the axial plots the diffusion profile calculated
    by the function :func:`poreana.diffusion.bins`.

    Parameters
    ----------
    data : string
        Data dictionary from function func:`poreana.diffusion.bins`
    intent : string, optional
        Set to **plot**, for plotting or set to **line** to only return the
        lineplot, leave empty for nothing
    kwargs: dict, optional
        Dictionary with plotting parameters
    """
    # Initialize
    width = data["width"]
    diff = data["diff"]
    is_norm = data["is_norm"]

    # Normalize x-axis
    if is_norm:
        for i in range(len(diff)-1, 0, -1):
            if diff[i] > 0:
                x_max = width[i+1]
                break

        bins_norm = [x/x_max for x in width]

    # Plot
    if intent == "plot" or intent == "line":
        x_axis = bins_norm if is_norm else width
        sns.lineplot(x=x_axis[:-1], y=diff, **kwargs)

    if intent == "plot":
        if is_norm:
            plt.xlabel("Normalized distance from pore center")
        else:
            plt.xlabel("Distance from pore center (nm)")
        plt.ylabel(r"Diffusion coefficient ($10^{-9}$ m${^2}$ s$^{-1}$)")


def mean(diff_data, dens_data, ax_area=[0.2, 0.8], is_print=True):
    """This function uses the diffusion coefficient slope obtained from
    function :func:`poreana.diffusion.bins` and the density slope of function
    :func:`poreana.density.bins` to calculate a weighted diffusion
    coefficient inside the pore

    .. math::

        \\langle D_\\text{axial}\\rangle
        =\\frac{\\int\\rho(r)D_\\text{axial}(r)dA(r)}{\\int\\rho(r)dA(r)}.

    In a discrete form, following formula is evaluated

    .. math::

        \\langle D_\\text{axial}\\rangle=\\frac{\\sum_{i=1}^n\\rho(r_i)D_\\text{axial}(r_i)A(r_i)}{\\sum_{i=1}^n\\rho(r_i)A(r_i)}

    with the partial area

    .. math::

        A(r_i)=\\pi(r_i^2-r_{i-1}^2)

    of radial bin :math:`i`. It is assumed that the discretization of the
    density is finer than the diffusion. Therefore, the diffusion values for
    each density bin are interpolated between the nearest available diffusion
    values.

    Parameters
    ----------
    diff_data : dictionary
        Diffusion data dictionary from function :func:`poreana.diffusion.bins`
    dens_data : dictionary
        Density data dictionary from function :func:`poreana.density.bins`
    ax_area : list, optional
        Bin area percentage to calculate the axial diffusion coefficient
    is_print : bool, optional
        True to print mean diffusion

    Returns
    -------
    diff_weight : float
        Density weighted mean axial diffusion in 10^-9 m^2s^-1
    """
    # Load density
    dens_x = dens_data["sample"]["data"]["in_width"]
    dens_y = dens_data["num_dens"]["in"]

    # Load diffusion
    diff_w = diff_data["width"][1]
    diff_x = diff_data["width"]
    diff_y = diff_data["diff"]

    # Integrate density and density weighted diffusion
    temp_diff = []
    dens_int = 0
    diff_int = 0
    for dens_bin, dens_x_val in enumerate(dens_x[:-1]):
        # only consider effective radius
        if dens_y[dens_bin]>0:
            # Find closest bins in diffusion
            diff_bins = [math.floor(dens_x_val/diff_w), math.ceil(dens_x_val/diff_w)]

            # Check boundary
            if diff_bins[1] <= len(diff_x):
                interp_x = [diff_x[bin_id] for bin_id in diff_bins]
                interp_y = [diff_y[bin_id] for bin_id in diff_bins]

                # Linear interpolation for diffusion
                diff_y_val = np.interp(dens_x_val, interp_x, interp_y)
                temp_diff.append(diff_y_val)
            else:
                diff_y_val = diff_y[diff_bins[0]]

            # Integrate
            int_area = (dens_x[dens_bin+1]**2-dens_x[dens_bin]**2)
            dens_int += int_area*dens_y[dens_bin]
            diff_int += int_area*dens_y[dens_bin]*diff_y_val

    # Normalize
    diff_weight = diff_int/dens_int

    # Output
    if is_print:
        print("Mean Diffusion axial: "+"%.3f" % diff_weight+" 10^-9 m^2s^-1")

    return diff_weight


######################################
# Diffusion - MC - Transition Matrix #
######################################
def mc_trans_mat(link, step, kwargs={}, is_norm=False, is_diagonal=False, is_length=True, limit = 0):
    """This function plots the occupation of a normalized transition matrix as a
    heatmap. To normalize the transition matrix the number of the frame are used.
    This means that all entries in the transition matrix are divided by the
    total number of frames which are used to estimate the matrix.

    Parameters
    ----------
    link: string
        Link to the diffusion hdf5, obj or yml data file generated by the sample routine
        :func:`poreana.sample.Sample.init_diffusion_mc` or the MC run
        :func:`poreana.mc.MC.run`
    step: integer
        Step length for which the transition matrix is to be displayed
    kwargs: dict, optional
        Dictionary with plotting parameters
    is_norm: bool,optional
        Normalized the transition with the number for frames
    is_diagonal: bool, optional
        Set the matrix diagonal to zero
    is_length: bool, optional
        x axis is output in box length
    limit: float, optional
        Set the lowest occupancy value for calculating the width of occupancy
    """
    # Load results from the output object file
    data = utils.load(link)

    # Sample obj file is loaded
    if "data" in data:
        trans_mat = data["data"][step]
        inp = data["inp"]
        frame_num = inp["num_frame"]
        frame_length = float(inp["len_frame"])
        bins = inp["bins"]

    # MC obj file is loaded
    else:
        model = data["model"]
        trans_mat = model["data"][step]
        frame_num = model["num_frame"]
        frame_length = float(model["len_frame"])
        bins = model["bins"]

    # Calculation of the width of the diagonal occupation of the transition matrix
    results=[]
    # Loop over the rows of the transition matrix
    for i in range(len(trans_mat[:])-1):
        # Right side of the matrix (row)
        row_right= trans_mat[i][i:]/frame_num
        # Left side of the matrix (row)
        row_left = trans_mat[i][:i]/frame_num
        # If a one element is zero call index
        if np.where(row_right<=limit)[0].size != 0:
            results.append(np.min(np.where(row_right<=limit)))
        if np.where(row_left<=limit)[0].size != 0:
            results.append(i-np.max(np.where(row_left<=limit)))

    # Get average of index
    idx_avg = np.mean(results)
    print("Width of occupancy: " + str(idx_avg))


    # Normalized transition matrix with frame number
    if is_norm:
        trans_mat = trans_mat/frame_num

    # Set diagonal elements of the transition matrix to zero
    if is_diagonal:
        trans_mat[np.diag_indices_from(trans_mat)] = 0



    # Set x axis from bins to box length
    if is_length:
        trans_new={}
        for i in range(len(bins)-1):
            trans_new[str(round(bins[i],2))] = trans_mat[i]
        trans_mat = pd.DataFrame(trans_new)


    # Set title with selected lag time
    plt.title("Lagtime: "+ str(step * frame_length * 10**12) + " ps", fontsize=10)

    # Plot the normalized transition matrix in a heatmap
    sns.heatmap(trans_mat, **kwargs)
    if is_length:
        plt.xlabel("Box length (nm)")


##################
# Diffusion - MC #
##################
def mc_fit(link, len_step=[], section=[], is_std=True, is_print=True, is_plot=True, kwargs_scatter={}, kwargs_line={}):
    """This function uses the diffusion profiles over box length which are
    calculated in the function :func:`poreana.mc.MC.run` to estimate
    the final diffusion coefficient. For that a line is fitted of the averaged
    diffusion profiles :math:`D_{\\mathrm{mean}}({\\Delta t_{\\alpha}})` for the
    different lag times as a function of :math:`1/\\Delta t_{\\alpha}`. The
    function plots all mean diffusion
    :math:`D_\\mathrm{mean}({\\Delta t_{\\alpha}})` over the
    inverse lag times and the linear fit. Additionally, the final diffusion
    coefficient :math:`D` for a lag time
    :math:`\\Delta t_{\\alpha} \\rightarrow \infty` is printed. Also, a table
    of the selected step length and the belonging
    :math:`D_{\\mathrm{mean}}({\\Delta t_{\\alpha}})` can be displayed.

    In order to be able to estimate how the choice of lag times affects the
    result, an additional error estimation is possible. For this all possible
    fourth tuples of the entire calculated lag times are used to estimate a
    mean diffusion coefficient :math:`\\langle D \\rangle` which is the mean
    value of all fourth tuples fitting results. Also, the standard deviation of
    all fitting results is printed. If there is no big difference between the
    diffusion coefficient :math:`D` and the mean diffusion coefficient
    :math:`\\langle D \\rangle`, all :math:`D_\\mathrm{mean}({\\Delta t_{\\alpha}})`
    are on a straight line. The standard deviation also can be used to check
    if the result fluctuates strongly at the choice of other lag times. This
    error estimate is calculated if :math:`\\mathrm{is}\_{\\mathrm{std}}=True`.

    Parameters
    ----------
    link: string
        Link to the diffusion hdf5 data file generated by the MC Alogrithm
        :func:`poreana.mc.MC.run`
    len_step: integer, list, optional
        List of the different step length, if it is [] all calculated lag times
        will be used to fit the diffusion coefficent
    section : list, string, optional
        If :math:`\\mathrm{section} = \\mathrm{"pore"}` the pore section is
        fitted.
        If :math:`\\mathrm{section} = \\mathrm{"reservoir"}` the left reservoir
        area is fitted.
        If :math:`\\mathrm{section} = \\mathrm{[a,b]}` the box area between a
        and b is fitted.
    is_std : bool, optional
        True to calculate a mean diffusion coefficient
        to assess the dependency of the result on the selected lag time
    is_print : bool, optional
        Print diffusion coefficient
    is_plot : bool, optional
        Show the fitting plot
    kwargs_scatter: dict, optional
        Dictionary with plotting parameters for the points
    kwargs_line: dict, optional
        Dictionary with plotting parameters for the fitting line


    Returns
    -------
    diffusion: float
        Diffusion coefficient :math:`D \ \\left( 10^{-9} \\frac{m^2}{s}\\right)` for the
        calculated system
    diffusion_mean: float
        Mean diffusion coefficient :math:`\\langle D \\rangle \ \\left(10^{-9} \\frac{m^2}{s}\\right)`
        for the calculated system
    diff_table : obj
        Table of used lag times with the associated
        :math:`D_{\\mathrm{mean}}({\\Delta t_{\\alpha}}) \ \\left(10^{-9} \\frac{m^2}{s}\\right)`
    res : float
        residual :math:`\\left(10^{-9} \\frac{m^2}{s}\\right)` for fitting the selected lag times
    """
    # Load data
    data = utils.load(link)


    results = data["output"]
    diff_bin = results["diff_profile"]
    model = data["model"]
    bins = model["bins"]
    dt = model["len_frame"]
    diff_unit = model["diffusion unit"]

    # If no specific step length is chosen take the step length from the object file
    if not len_step:
        len_step = model["len_step"]

    # If a pore system is considered
    if "pore" in data:
        pore = data["pore"]
        res = float(pore["res"])
        box = pore["box"]

    # Set vector
    diff_bin_vec = {}

    # Pore
    if isinstance(section, str) and section== "pore":
        # If only the pore area should be considered
        if "pore" in data:
            # Set section
            area = [res, box[2]-res]

        # If only the pore area should be considered
        else:
            print("obj.-file includes results of a simple box")
            return;

    # Reservoir
    elif isinstance(section, str) and section == "reservoir":
        # If only the pore area should be considered
        # Load pore system
        if "pore" in data:
            # Set section
            area = [0,res]

        # If only the pore area should be considered
        else:
            print("obj.-file includes results of a simple box")
            return;

    elif isinstance(section, str) and section not in ["reservoir", "pore"]:
        print("Wrong input for section! Check documentation for available options")
        return;

    elif isinstance(section, list) and len(section)>=3:
        print("Wrong input for section! Check documentation for available options")
        return;

    #Area section
    elif isinstance(section, list):
        area = section

    # If section is not defined -> entire system
    if not section:
        # Set section
        area = [bins[0], bins[-1]]

    # Cut profile
    # Calculated start and end bin index of the pore area
    index_start = np.digitize(area[0], bins)
    index_end = np.digitize(area[1], bins)


    for i in len_step:
        diff_bin_vec[i] = [diff_bin[i][j] for j in range(index_start, index_end)]

    # Calculate mean diffusion coefficient and standard deviation
    # Mean diffusion means the averaged of all possible fourth tuple fitting results
    if is_std:
        len_step_all = len_step

        # Determine all possible combinations
        a = list(itertools.combinations(len_step_all, 2))

        # Allocate the result vector
        res = np.zeros(len(a))

        # Fit to infinity lag time for all combinations
        for i in range(len(a)):

            # Set the current tuple
            rand = list(a[i])

            # Calculate the mean diffusion (m^2/s) over all bins
            D_mean = [np.mean(np.exp([diff_bin_vec[i][j] + diff_unit for j in range(len(diff_bin_vec[i]))])) * 10**3 for i in rand]


            # Calculate the inverse lag time for the linear fit
            lagtime_inverse = [1 / (i * dt * 10**-12) for i in rand]

            # Fit a linear line
            fit = np.poly1d(np.polyfit(lagtime_inverse, D_mean, 1))

            # Set the fitted diffusion coefficient on the results list
            res[i] = fit(0)

        # Determine standard deviation and mean diffusion of fourth touple method
        diffusion_mean = np.mean(res)
        std = res.std()

    # Calculate the mean diffusion (m^2/s) over all bins
    D_mean = [np.mean(np.exp([diff_bin_vec[i][j] + diff_unit for j in range(len(diff_bin_vec[i]))])) * 10**3 for i in len_step]

    # Calculate the inverse lag time (1/s) for the linear fit
    lagtime_inverse = [1 / (len_step[i] * dt * 10**-12) for i in range(len(len_step))]

    # Fit a linear line
    fit = np.poly1d(np.polyfit(lagtime_inverse, D_mean, 1))

    # Calculate residuals
    res = np.polyfit(lagtime_inverse, D_mean, 1, full=True)

    # Print the diffusion coefficient for an entire system
    if is_print:
        if not section:
            print("\nDiffusion axial: "+"%.4e" % (fit(0) * 10 **-9) + " m^2/s\n")

            # Print resudial for fitting
            print("Residual: "+"%.4e" % (float(res[1]) * 10 **-9) + " m^2/s\n")

            # If is_std true print the results of the calculations
            if is_std:
                print("Mean Diffusion axial: "+"%.4e" % (diffusion_mean * 10 **-9)  + " m^2/s\n")

        # Print the diffusion coefficient in the pore area
        if section=="pore":
            print("\nDiffusion axial (Pore): "+"%.4e" % (fit(0) * 10 **-9) + " m^2/s\n")


            # Print resudial for fitting
            print("Residual: "+"%.4e" % (float(res[1]) * 10 **-9) + " m^2/s\n")

            # If is_std true print the results of the calculations
            if is_std:
                print("Mean Diffusion axial (Pore): "+"%.4e" % (diffusion_mean * 10 **-9)  + " m^2/s\n")

        # Print the diffusion coefficient in the reservoir area
        if section=="reservoir":
            print("\nDiffusion axial (Reservoir): "+"%.4e" % (fit(0) * 10 **-9) + " m^2/s\n")

            # Print resudial for fitting
            print("Residual: "+"%.4e" % (float(res[1]) * 10 **-9) + " m^2/s\n")

            # If is_std true print the results of the calculations
            if is_std:
                print("Mean Diffusion axial (Peservoir): "+"%.4e" % (diffusion_mean * 10 **-9) + " m^2/s\n")

        # Print the diffusion coefficient in a selected section
        if (isinstance(section, list)) and len(section)==2:
            print("\nDiffusion axial ([" + "%.2f" % (area[0]) + ", " + "%.2f" % (area[1]) + "]): "+"%.4e" % (fit(0) * 10 **-9) + " m^2/s\n")

            # Print resudial for fitting
            print("Residual: "+"%.4e" % (float(res[1]) * 10 **-9) + " m^2/s\n")

            # If is_std true print the results of the calculations
            if is_std:
                print("Mean Diffusion axial (["+ "%.2f" % (area[0]) + ", " + "%.2f" % (area[1]) +"]): "+"%.4e" % (diffusion_mean * 10 **-9)  + " m^2/s\n")

        # Print std deviation
        if is_std:
            print("Standard deviation: "+"%.4e" % (std * 10 **-9) + " m^2/s\n")

    # Set data frame for the used lag times
    data = [str("%.2f" % D_mean[i] ) for i in range(len(len_step))]
    diff_table = pd.DataFrame(data, index=list(len_step), columns=list(['$D_\mathrm{mean} \ (10^{-9} \mathrm{m^2s^{-1}})$']))
    diff_table = pd.DataFrame(diff_table.rename_axis('Step Length', axis=1))
    styler = diff_table.style.set_caption('Selected step length')
    diff_table = styler.set_properties(**{'text-align': 'right'})
    diff_table = diff_table.set_table_styles([dict(selector='th', props=[('text-align', 'left')])])

    # Set vectors for plotting
    D_mean_vec = [D_mean[i] for i in range(len(lagtime_inverse))]
    lag_time_vec = [1 / (len_step[i] * dt) for i in range(len(len_step))]
    x_vec = np.arange(0, max(lag_time_vec) * 2, (max(lag_time_vec) * 2) / 5)

    # Fit a linear line and calculated diffusion coefficent
    fit = np.poly1d(np.polyfit(lag_time_vec, D_mean_vec, 1))
    diffusion = fit(0)

    if is_plot:
        # Plot the results
        plt.xlim(0, 1.5*max(lag_time_vec))
        plt.ylim(0, 1.5*max(fit(x_vec)))
        sns.scatterplot(x=lag_time_vec, y=D_mean_vec, **kwargs_scatter)
        sns.lineplot(x=x_vec, y=fit(x_vec), **kwargs_line)
        legend = ["$D_{\mathrm{fit}}$", "$D_{\mathrm{mean}}(\\Delta t_{\\alpha})$"]
        plt.legend(legend)
        plt.xlabel(r"Inverse lag time ($10^{12} \ \mathrm{s^{-1}})$")
        plt.ylabel(r"Diff. coeff. ($10^{-9} \ \mathrm{m^2s^{-1}}$)")

    return diffusion, diffusion_mean, diff_table, float(res[1][0])



def mc_profile(link, len_step=[], section=[], infty_profile=True,  is_plot=True, kwargs={}):
    """This function plots the diffusion profile for an infinity
    lag time (:math:`\\Delta t_{\\alpha} \\rightarrow \\infty`) over the box
    fitted with the specified :math:`\\mathrm{len}\_\\mathrm{step}` list.
    Additionally, it is possible to display the diffusion profiles for the
    calculated lag times. Therefore, the
    :math:`\\mathrm{infty}\_\\mathrm{profile}` has to be false.
    The list :math:`\\mathrm{len}\_\\mathrm{step}`
    contains the calculated step length which should be
    displayed.

    Parameters
    ----------
    link : string
        Link to the diffusion hdf5 data file generated by the MC Algorithm
        :func:`poreana.mc.MC.run`
    len_step : integer list, optional
        List of the different step length, if it is [] all diffusion profiles
        depending on the lag time are used to calculate the diffusion profile
        for an infinite lag time or the individual diffusion profiles for these
        lag times are shown
    section : list, string, optional
        If :math:`\\mathrm{section} = \\mathrm{"pore"}` the pore section is
        fitted.
        If :math:`\\mathrm{section} = \\mathrm{"reservoir"}` the left reservoir
        area is fitted.
        If :math:`\\mathrm{section} = \\mathrm{[a,b]}` the box area between a
        and b is fitted.
    infty_profile : bool, optional
        Set to false to display all individual diffusion profiles for
        the selected lag times
    is_plot : bool, optional
        Show diffusion profile
    kwargs: dict, optional
        Dictionary with plotting parameters

    Returns
    -------
    diff_profile_fit : list
        Diffusion profile :math:`\\left(10^{-9} \\frac{m^2}{s}\\right)` for an infinite
        lag time
    diff_profiles : dict
        Diffusion profiles :math:`\\left(10^{-9} \\frac{m^2}{s}\\right)` for every
        calculated lag time
    bins : list
        bins over the box length
    res : float
        residual :math:`\\left(10^{-9} \\frac{m^2}{s}\\right)` for fitting the infinite diffusion profile
    """
    # Load data
    data = utils.load(link)

    # Load results
    results = data["output"]
    diff_bin = results["diff_profile"]

    # Load model inputs
    model = data["model"]
    diff_unit = model["diffusion unit"]
    bins = model["bins"]
    dt = model["len_frame"]

    # If no specific step length is chosen take the step length from the object file
    if not len_step:
        len_step = model["len_step"]

    # If a pore system is considered
    if "pore" in data:
        pore = data["pore"]
        res = pore["res"]
        box = pore["box"]


    # Set dictionaries
    legend = []
    diff_bin_vec = {}
    diff_profile_fit = []
    res_list = []


    # Pore
    if isinstance(section, str) and section== "pore":
        # If only the pore area should be considered
        # Load pore system
        if "pore" in data:
            # Set section
            area = [res, box[2]-res-2*(bins[1]-bins[0])]


        # If only the pore area should be considered
        else:
            print("obj.-file includes results of a simple box")
            return;

    # Reservoir
    elif isinstance(section, str) and section== "reservoir":
        # If only the pore area should be considered
        # Load pore system
        if "pore" in data:
            # Set section
            area = [0,res]

        # If only the pore area should be considered
        else:
            print("obj.-file includes results of a simple box")
            return;

    elif isinstance(section, str) and section not in ["reservoir", "pore"]:
        print("Wrong input for section! Check documentation for available options")
        return;

    elif isinstance(section, list) and len(section)>=3:
        print("Wrong input for section! Check documentation for available options")
        return;

    #Area section
    elif isinstance(section, list):
        area = section

    # If section is not defined -> entire system
    if not section:
        # Set section
        area = [bins[0], bins[-1]]

    # Cut profile
    # Calculated start and end bin index of the pore area
    index_start = np.digitize(area[0], bins)
    index_end = np.digitize(area[1], bins)

    # Save for all lag times the cutted profile
    for i in len_step:
        diff_bin_vec[i] = [diff_bin[i][j] for j in range(index_start, index_end)]

    # Set bin list
    bins = [bins[i] for i in range(index_start, index_end)]


    # Calculate the inverse lag time (1/s) for the linear fit
    lagtime_inverse = [1 / (len_step[i] * dt * 10**-12) for i in range(len(len_step))]

    # Calculate diffusion profiles
    diff_profiles = {}
    for i in len_step:
        diff_profiles[i] = [np.exp(diff_bin_vec[i][j] + diff_unit) * 10 ** 3 for j in range(len(bins))]

    # If infty_profile is false the profiles for the different lag times are plotted

    if not infty_profile:
        # Plot the profiles for the
        if is_plot:
            for i in len_step:
                sns.lineplot(x=bins, y=(diff_profiles[i]), **kwargs)       # Diffusion in m^2/s

        # Plot the diffusion profiles for the different lag times
        legend = ["$\\Delta t_{\\alpha}$ = " + str(len_step[i] * dt) + " ps" for i in range(len(len_step))]

    # If infty_profile is True the diffusion profile for a infinity lag times is shwon
    if len(len_step) >= 2 and infty_profile:
        # Initialize fit vector

        # Calculate the mean diffusion over all bins
        for i in range(len(bins)):
            diff = [np.exp(diff_bin_vec[step][i] + diff_unit) * 10 ** 3 for step in len_step]   # Diffusion in m^2/s

            # fit a linear line
            fit = np.poly1d(np.polyfit(lagtime_inverse, diff, 1))

            # Append diffusion at t-> infty
            diff_profile_fit.append(fit(0))

            # Calculate residuals for fitting
            res = np.polyfit(lagtime_inverse, diff, 1, full=True)

            # Append residual
            res_list.append(float(res[1]))

        # Plot fitted diffusion profile
        if is_plot:
            sns.lineplot(x=bins, y=diff_profile_fit, **kwargs)       # Diffusion in m^2/s

    # Set legend for lag times
    if is_plot:
        if not infty_profile and len(len_step) >= 2:
            legend.append("$\\Delta t_{\\alpha} \\rightarrow \\infty$ ps")


        # Set plot properties
        # Plot axis title for a entire system
        plt.ylabel(r"Diff. coeff. ($10^{-9} \ \mathrm{m^2s^{-1}}$)")
        plt.xlabel(r"Box length (nm)")
        plt.xlim([min(bins),max(bins)])
        if legend:
            plt.legend(legend)

    # Check if profile was fitted
    if not res_list:
        res_list = 0
    else:
        np.mean(res_list)

    return diff_profile_fit, diff_profiles, bins, res_list




###########################
# Diffusion - MC - Radial #
###########################
# def mc_fit_radial(link, len_step=None):
#     """This function uses the diffusion profiles over box length which are
#     calculated in the function :func:`do_mc_cycles` to estimate the final
#     diffusion coefficient. For that a line is fitted of the averaged diffusion
#     profiles :math:`D_{\\mathrm{mean}}({\\Delta_{ij}t_{\\alpha}})` for the
#     different lag times as a function of :math:`1/\\Delta_{ij}t_{\\alpha}`. The
#     function plots all mean diffusion
#     :math:`D_{mean}({\\Delta_{ij}t_{\\alpha}})` over the inverse lag times and
#     the linear fit. Additional the final mean diffusion coefficient :math:`D`
#     for a lagtime :math:`\\Delta_{ij}t_{\\alpha} \\rightarrow \infty` is
#     printed. Also a table of the selected step length and the belonging
#     :math:`D_{\\mathrm{mean}}({\\Delta_{ij}t_{\\alpha}})` can be displayed.
#
#     In order to be able to estimate how the choice of lag times affects the
#     result, an additional error estimation is possible. For this all possible
#     fourth tuples of the entire calculated lag times are used to estimate a
#     mean diffusion coefficient :math:`\\langle D \\rangle` which is the mean
#     value of all fourth tuples fitting results. Also the standard deviation of
#     all fitting results is printed. If there is no big difference between the
#     diffusion coefficient :math:`D` and the mean diffusion ceofficient
#     :math:`\\langle D \\rangle`, all :math:`D_{mean}({\\Delta_{ij}t_{\\alpha}})`
#     are on a straight line. The standard deviation can also be used to check
#     if the result fluctuates strongly at the choice of other lag times. This
#     error estimate is calculated if :math:`\\mathrm{is}_{\\mathrm{std}}=True`.
#
#     Parameters
#     ----------
#     link: string
#         Link to the diffusion data object generated by the sample routine
#         :func:`MC.do_mc_cycles`
#     len_step: integer list, optional
#         List of the different step length, if it is None the lag times which
#         are contains in the obj file are printed
#     is_std : bool, optional
#         True to calculate a mean diffusion coefficient and a Bootstrap diffusion
#         to assess the dependency of the results on the chosen lag time
#
#     Returns
#     -------
#     diffusion: float
#         Radial diffusion coefficient :math:`D \ \\left(\\frac{m^2}{s}\\right)` for the
#         calculated system
#     diffusion_mean: float
#         Radial mean diffusion coefficient :math:`\\langle D \\rangle \ \\left(\\frac{m^2}{s}\\right)` for the
#         calculated system
#     diff_table : obj
#         Table of used lag times with the associated
#         :math:`D_{\\mathrm{mean}}({\\Delta_{ij}t_{\\alpha}}) \ \\left(\\frac{m^2}{s}\\right)`
#     """
#     # Load Results from the output object file
#     results = utils.load(link)
#     diff_bin = results["diff_radial_profile"]
#     df_bin = results["df_profile"]
#     inp = results["model"]
#     bins = inp["bins"]
#     diff_unit = inp["diffusion radial unit"]
#     dt = inp["len_frame"]
#
#     # If no specific step length is chosen take the step length from the object file
#     if not len_step:
#         len_step = inp["len_step"]
#
#     # Calculate the mean diffusion over all bins
#     D_mean = [np.mean(np.exp(diff_bin[i] + diff_unit)) * 10**-
#               6 for i in len_step]   # Diffusion in m^2/s
#
#     # Calculate the inverse lag time (1/s) for the linear fit
#     lagtime_inverse = [1 / (len_step[i] * dt * 10**-12) for i in range(len(len_step))]
#
#     # Fit a linear line
#     fit = np.poly1d(np.polyfit(lagtime_inverse, D_mean, 1))
#
#     # set the x value for the linear fit
#     x = np.arange(0, max(lagtime_inverse) * 2, (max(lagtime_inverse) * 2) / 5)
#
#     # Calculate the diffusion coefficient as a mean value over the bin
#     print("\nMean Radial Diffusion: "+"%.4e" % fit(0) + " m^2/s\n")
#
#     # Integration
#     bin_width = bins[1]-bins[0]
#
#     # Set data frame for the used lag times
#     D_mean_pd = [str("%.2e" % D_mean[i]) for i in range(len(D_mean))]
#     diff_table = pd.DataFrame({'$D_\mathrm{mean}$': D_mean_pd, '$D_\mathrm{avg} (\mathrm{m^2s^{-1}})$': diff_avg_pd}, index=list(len_step))
#     diff_table = pd.DataFrame(diff_table.rename_axis('Step Length', axis=1))
#     styler = diff_table.style.set_caption('Selected step length')
#     diff_table = styler.set_properties(**{'text-align': 'right'})
#     diff_table = diff_table.set_table_styles([dict(selector='th', props=[('text-align', 'left')])])
#
#     # Plot the results
#     plt.xlim(0, 1.5*max(lagtime_inverse))
#     plt.ylim(0, 1.5*max(fit(x)))
#     sns.scatterplot(x=lagtime_inverse, y=D_mean, color='red')
#     sns.lineplot(x=x, y=fit(x))
#     legend = ["linear fit", "$D_{\mathrm{mean}}$"]
#     plt.legend(legend)
#     plt.xlabel(r"Inverse lagtime ($\mathrm{s^{-1}})$")
#     plt.ylabel(r"radial Diff. coeff. ($\mathrm{m^2s^{-1}}$)")
#
#     return fit(0), diff_table
#
#
# def mc_profile_radial(link, len_step=None, avg=False):
#     """This function plots the radial diffusion profile for a infinity
#     lag time (:math:`\\Delta_{ij}t_{\\alpha} \\rightarrow \\infty`) over the box
#     fitted with the specified :math:`\\mathrm{len}\_\\mathrm{step}` list.
#     Additionally, it is possible to display the diffusion profiles for the
#     calculated lag times. Therefore, the
#     :math:`\\mathrm{infty}\_\\mathrm{profile}` has to be false.
#     The list :math:`\\mathrm{len}\_\\mathrm{step}`
#     contains the calculated step length which should be displayed.
#
#     Parameters
#     ----------
#     link : string
#         Link to the diffusion data object generated by the
#         :func:`mc.MC.do_mc_cycles`
#     len_step : integer list, optional
#         List of the different step length, if it is None the lag times which
#         are contains in the obj file are printed
#     infty_profile : bool, optional
#         If its True only the profile for an infity lagtime is plotted
#     """
#     # Load Results from the output object file
#     results = utils.load(link)
#     diff_bin = results["diff_radial_profile"]
#     inp = results["model"]
#     dt = inp["len_frame"]
#     diff_unit = inp["diffusion radial unit"]
#     bins = inp["bins"]
#     bins = [(bins[i]) * 10**-8 for i in range(len(bins))]
#
#     if len_step is None:
#         len_step = inp["len_step"]
#
#     # Plot the diffusion profiles for the different lag times
#     legend = ["lagtime " + str(len_step[i] * dt) + " ps" for i in range(len(len_step))]
#
#     # Plot the diffusion profiles (m^2/s) for the
#     for i in len_step:
#         sns.lineplot(x=bins, y=np.exp(diff_bin[i] + diff_unit) * 10**-6)
#
#     # Averged profile over the selected lag times
#     if len(len_step) != 1 and avg:
#         diff_mean = np.zeros(len(bins[:-1]))
#         diff = []
#         for i in len_step:
#             diff_mean += diff_bin[i]
#         diff = [diff_mean[i]/len(len_step) for i in range(len(diff_mean))]
#
#         # Plot mean profile (m^2/s) over the selected lag times
#         sns.lineplot(x=bins, y=np.exp(diff + diff_unit) * 10**-6)
#         legend.append("avg profile")
#
#     # Set plot properties
#     plt.xlabel(r"Box length (m)")
#     plt.ylabel(r"radial Diffusion ($\mathrm{m^2s^{-1}}$)")
#     plt.legend(legend)



import os
import sys

import copy
import shutil
import unittest

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

import porems as pms
import poreana as pa


class UserModelCase(unittest.TestCase):
    #################
    # Remove Output #
    #################
    @classmethod
    def setUpClass(self):
        if os.path.isdir("tests"):
            os.chdir("tests")

        folder = 'output'
        pa.utils.mkdirp(folder)
        pa.utils.mkdirp(folder+"/temp")
        open(folder+"/temp.txt", 'a').close()

        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)

        # Load molecule
        mol_B = pms.Molecule(inp="data/benzene.gro")
        mol_W = pms.Molecule(inp="data/spc216.gro")
        mol_H = pms.Molecule(inp="data/heptane.gro")

        #Sample
        # Single core
        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol_B)
        sample.init_density("output/dens_cyl_s.h5")
        sample.init_gyration("output/gyr_cyl_s.h5")
        sample.init_diffusion_bin("output/diff_cyl_s.h5")
        sample.sample(is_parallel=False)

        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol_B)
        sample.init_density("output/dens_cyl_no_remove.h5", remove_pore_from_res=False)
        sample.sample(is_parallel=False)

        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol_B)
        sample.init_density("output/dens_cyl_no_remove.obj", remove_pore_from_res=False)
        sample.sample(is_parallel=False)

        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol_B)
        sample.init_density("output/dens_cyl_no_remove.obj", remove_pore_from_res=False)
        sample.sample(is_parallel=False)

        sample = pa.Sample("data/pore_system_slit.yml", "data/traj_slit.xtc", mol_W)
        sample.init_density("output/dens_slit.h5")
        sample.sample(is_parallel=False, is_pbc=False)

        sample = pa.Sample([6.00035, 6.00035, 19.09191], "data/traj_box.xtc", mol_H)
        sample.init_density("output/dens_box.h5")
        sample.init_gyration("output/gyr_box.h5")
        sample.sample(shift=[0, 0, 3.3], is_parallel=False)

        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol_B)
        sample.init_diffusion_mc("output/diff_mc_cyl_s.h5", len_step=[1, 2, 5, 10, 20, 30, 40, 50, 100, 200, 250, 300, 350])
        sample.sample(is_parallel=False)
        #
        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol_B)
        sample.init_diffusion_mc("output/diff_mc_cyl_s.h5", len_step=[1, 2, 5, 10, 20, 30, 40, 50, 100, 200, 250, 300, 350])
        sample.sample(is_parallel=True)

        sample = pa.Sample([6.00035, 6.00035, 19.09191], "data/traj_box.xtc", mol_H)
        sample.init_diffusion_mc("output/diff_mc_box.h5", len_step=[1, 2, 5, 10, 20, 30, 40, 50])
        sample.sample(shift=[0, 0, 3.3], is_parallel=False, is_pbc=True)
        #
        # ## Parallel
        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol_B)
        sample.init_density("output/dens_cyl_p.h5")
        sample.init_gyration("output/gyr_cyl_p.h5")
        sample.init_diffusion_bin("output/diff_cyl_p.h5")
        sample.sample(is_parallel=True, is_pbc=False)

        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol_B)
        sample.init_diffusion_mc("output/diff_mc_cyl_p.h5", len_step=[1, 2, 5, 10, 20, 30, 40, 50, 100, 200, 250, 300, 350])
        sample.sample(is_parallel=True, is_pbc=True, np=6)


    #########
    # Utils #
    #########
    def test_utils(self):
        file_link = "output/test/test.obj"

        pa.utils.mkdirp("output/test")

        self.assertEqual(pa.utils.column([[1, 1, 1], [2, 2, 2]]), [[1, 2], [1, 2], [1, 2]])

        pa.utils.save([1, 1, 1], file_link)
        self.assertIsNone(pa.utils.load("output/test/test.txt", file_type="DOTA"))
        pa.utils.file_to_text("data/box_output.h5", "output/box_output.txt")
        pa.utils.file_to_text("data/check_output.h5", "output/check_output.txt")
        pa.utils.file_to_text("output/dens_cyl_s.h5", "output/dens_cyl_s.txt")
        pa.utils.file_to_text("output/diff_cyl_s.h5", "output/diff_cyl_s.txt", "output/dens_cyl_s.h5",)
        pa.utils.file_to_text("output/diff_cyl_s.h5", "output/diff_cyl_s.txt")
        pa.utils.file_to_text("output/gyr_cyl_s.h5", "output/gyr_cyl_s.txt", "output/dens_cyl_s.h5",)
        pa.utils.file_to_text("output/gyr_cyl_s.h5", "output/gyr_cyl_s.txt")
        pa.utils.file_to_text("output/gyr_box.h5", "output/gyr_box.txt", "output/dens_box.h5",)
        pa.utils.file_to_text("output/dens_box.h5", "output/dens_box.txt")

        print(pa.utils.load(file_link))
        self.assertEqual(pa.utils.load(file_link), [1, 1, 1])

        self.assertEqual(round(pa.utils.mumol_m2_to_mols(3, 100), 4), 180.66)
        self.assertEqual(round(pa.utils.mols_to_mumol_m2(180, 100), 4), 2.989)
        self.assertEqual(round(pa.utils.mmol_g_to_mumol_m2(0.072, 512), 2), 0.14)
        self.assertEqual(round(pa.utils.mmol_l_to_mols(30, 1000), 4), 18.066)
        self.assertEqual(round(pa.utils.mols_to_mmol_l(18, 1000), 4), 29.8904)

        print()
        pa.utils.toc(pa.utils.tic(), message="Test", is_print=True)
        self.assertEqual(round(pa.utils.toc(pa.utils.tic(), is_print=True)), 0)

    ############
    # Geometry #
    ############
    def test_geometry(self):
        vec_a = [1, 1, 2]
        vec_b = [0, 3, 2]

        print()

        self.assertEqual(round(pa.geom.dot_product(vec_a, vec_b), 4), 7)
        self.assertEqual(round(pa.geom.length(vec_a), 4), 2.4495)
        self.assertEqual([round(x, 4) for x in pa.geom.vector(vec_a, vec_b)], [-1, 2, 0])
        self.assertIsNone(pa.geom.vector([0, 1], [0, 0, 0]))
        self.assertEqual([round(x, 4) for x in pa.geom.unit(vec_a)], [0.4082, 0.4082, 0.8165])
        self.assertEqual([round(x, 4) for x in pa.geom.cross_product(vec_a, vec_b)], [-4, -2, 3])
        self.assertEqual(round(pa.geom.angle(vec_a, vec_b), 4), 37.5714)

    ############
    # # Sample #
    # ##########
    def test_sample(self):
        # self.skipTest("Temporary")

        print()

        # Define molecules
        mol = pms.Molecule("benzene", "BEN", inp="data/benzene.gro")
        mol2 = pms.Molecule("benzene", "BEN", inp="data/benzene.gro")
        mol2.add("H", [0, 0, 0])

        # Sanity checks
        pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol2)
        pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol, atoms=["C1"], masses=[1, 1])
        pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol).sample(shift=[1])

        # Diffusion
        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol, atoms=["C1"])
        sample.init_diffusion_bin("output/diff_np_s.h5", len_obs=3e-12)

        sample = pa.Sample([0, 0, 1], "data/traj_cylinder.xtc", mol, atoms=["C1"])
        sample.init_diffusion_bin("output/diff_box_test.h5", len_obs=3e-12)


        # Test the error output if bin and MC diffusion calculation is initialized
        # Test bin -> MC
        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol)
        sample.init_diffusion_bin("output/test.h5")
        sample.init_diffusion_mc("output/test.h5", len_step=[1, 2, 5, 10])

        # Test MC -> Bin
        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol)
        sample.init_diffusion_mc("output/test.h5", len_step=[1, 2, 5, 10])
        sample.init_diffusion_bin("output/test.h5")

        # Test direction wrong input
        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol)
        sample.init_diffusion_mc("output/test.h5", len_step=[1, 2, 5, 10], direction=4)

        # Test direction wrong input
        sample = pa.Sample("data/pore_system_cylinder.yml", "data/traj_cylinder.xtc", mol)
        sample.init_diffusion_mc("output/test.obj", len_step=[1, 2, 5, 10], direction=4)

    ##############
    # Adsorption #
    ##############
    def test_adsorption(self):
        # self.skipTest("Temporary")

        # Calculate adsorption
        ads_s = pa.adsorption.calculate("output/dens_cyl_s.h5")
        ads_p = pa.adsorption.calculate("output/dens_cyl_p.h5")

        self.assertEqual(round(ads_s["conc"]["mumol_m2"], 2), 0.15)
        self.assertEqual(round(ads_s["num"]["in"], 2), 10.77)
        self.assertEqual(round(ads_p["conc"]["mumol_m2"], 2), 0.15)
        self.assertEqual(round(ads_p["num"]["in"], 2), 10.77)


    ###########
    # Density #
    ###########
    def test_density(self):
        # self.skipTest("Temporary")

        # Calculate density
        dens_s = pa.density.bins("output/dens_cyl_s.h5", target_dens=16)
        dens_p = pa.density.bins("output/dens_cyl_p.h5", target_dens=16)
        dens_s_mean = pa.density.mean(dens_s)

        dens_no_remove = pa.density.bins("output/dens_cyl_no_remove.obj")

        dens_slit = pa.density.bins("output/dens_slit.h5", target_dens=997)
        dens_box = pa.density.bins("output/dens_box.h5")

        # Plot density
        plt.figure()
        pa.density.bins_plot(dens_s, target_dens=0.146)
        pa.density.bins_plot(dens_s, target_dens=0.146, is_mean=True)
        plt.savefig("output/density.pdf", format="pdf", dpi=1000)
        # plt.show()

        plt.figure()
        pa.density.bins_plot(dens_no_remove)
        plt.savefig("output/density_no_remove.pdf", format="pdf", dpi=1000)
        # plt.show()

        plt.figure()
        pa.density.bins_plot(dens_slit, is_mean=True)
        plt.savefig("output/density_slit.pdf", format="pdf", dpi=1000)
        # plt.show()

        plt.figure()
        pa.density.bins_plot(dens_s, intent="in")
        pa.density.bins_plot(dens_p, intent="in")
        plt.legend(["Single core", "Parallel"])
        plt.savefig("output/density_in.pdf", format="pdf", dpi=1000)
        # plt.show()

        plt.figure()
        pa.density.bins_plot(dens_s, intent="ex")
        pa.density.bins_plot(dens_p, intent="ex")
        plt.legend(["Single core", "Parallel"])
        plt.savefig("output/density_ex.pdf", format="pdf", dpi=1000)
        # plt.show()

        plt.figure()
        pa.density.bins_plot(dens_box, intent="ex")
        plt.savefig("output/density_box.pdf", format="pdf", dpi=1000)
        # plt.show()

        # Run tests
        self.assertEqual(round(dens_s["dens"]["in"], 3), 12.982)
        self.assertEqual(round(dens_s["dens"]["ex"], 3), 16.432)
        self.assertEqual(round(dens_p["dens"]["in"], 3), 12.982)
        self.assertEqual(round(dens_p["dens"]["ex"], 3), 16.432)

        print()
        pa.density.bins_plot(dens_s, intent="DOTA")


    ############
    # Gyration #
    ############
    def test_gyration(self):
        # self.skipTest("Temporary")

        # Plot gyration radius
        plt.figure()
        mean_s = pa.gyration.bins_plot("output/gyr_cyl_s.h5", "output/dens_cyl_s.h5", is_mean=True)
        plt.savefig("output/gyration_s.pdf", format="pdf", dpi=1000)
        plt.figure()
        mean_p = pa.gyration.bins_plot("output/gyr_cyl_p.h5", "output/dens_cyl_p.h5")
        plt.savefig("output/gyration_p.pdf", format="pdf", dpi=1000)

        plt.figure()
        pa.gyration.bins_plot("output/gyr_cyl_s.h5", "output/dens_cyl_s.h5", intent="in")
        pa.gyration.bins_plot("output/gyr_cyl_p.h5", "output/dens_cyl_p.h5", intent="in")
        plt.legend(["Single core", "Parallel"])
        plt.savefig("output/gyration_in.pdf", format="pdf", dpi=1000)

        plt.figure()
        pa.gyration.bins_plot("output/gyr_cyl_s.h5", "output/dens_cyl_s.h5", intent="ex")
        pa.gyration.bins_plot("output/gyr_cyl_p.h5", "output/dens_cyl_p.h5", intent="ex")
        plt.legend(["Single core", "Parallel"])
        plt.savefig("output/gyration_ex.pdf", format="pdf", dpi=1000)

        plt.figure()
        pa.gyration.bins_plot("output/gyr_cyl_s.h5", "output/dens_cyl_s.h5", intent="in", is_norm=True)
        pa.gyration.bins_plot("output/gyr_cyl_s.h5", "output/dens_cyl_s.h5", intent="ex", is_norm=True)
        plt.legend(["Interior", "Exterior"])
        plt.savefig("output/gyration_norm.pdf", format="pdf", dpi=1000)

        plt.figure()
        pa.gyration.bins_plot("output/gyr_box.h5", "output/dens_box.h5", intent="ex")
        plt.savefig("output/gyration_box.pdf", format="pdf", dpi=1000)

        print()
        pa.gyration.bins_plot("output/gyr_cyl_s.h5", "output/dens_cyl_s.h5", intent="DOTA")

        self.assertEqual(round(mean_s["in"], 2), 0.13)
        self.assertEqual(round(mean_s["ex"], 2), 0.15)
        self.assertEqual(round(mean_p["in"], 2), 0.13)
        self.assertEqual(round(mean_p["ex"], 2), 0.15)

        self.assertEqual(round(mean_s["in"], 2), 0.13)
        self.assertEqual(round(mean_s["ex"], 2), 0.15)
        self.assertEqual(round(mean_p["in"], 2), 0.13)
        self.assertEqual(round(mean_p["ex"], 2), 0.15)

        self.assertEqual(round(mean_s["in"], 2), 0.13)
        self.assertEqual(round(mean_s["ex"], 2), 0.15)
        self.assertEqual(round(mean_p["in"], 2), 0.13)
        self.assertEqual(round(mean_p["ex"], 2), 0.15)


    #################
    # Bin Diffusion #
    #################
    def test_diffusion_bin(self):
        # self.skipTest("Temporary")

        # Bin diffusion
        plt.figure()
        pa.diffusion.bins_plot(pa.diffusion.bins("output/diff_cyl_s.h5"))
        pa.diffusion.bins_plot(pa.diffusion.bins("output/diff_cyl_s.h5", is_norm=True))
        plt.savefig("output/diffusion_bins.pdf", format="pdf", dpi=1000)
        # plt.show()

        # CUI diffusion
        plt.figure()
        pa.diffusion.cui("output/diff_cyl_s.h5", is_fit=True)
        plt.savefig("output/diffusion_cui.pdf", format="pdf", dpi=1000)
        # plt.show()

        # Mean diffusion based on bins
        plt.figure()
        mean_s = pa.diffusion.mean(pa.diffusion.bins("output/diff_cyl_s.h5"), pa.density.bins("output/dens_cyl_s.h5"))
        plt.savefig("output/diff_mean_check.pdf", format="pdf", dpi=1000)
        mean_p = pa.diffusion.mean(pa.diffusion.bins("output/diff_cyl_p.h5"), pa.density.bins("output/dens_cyl_p.h5"))

        self.assertEqual(round(mean_s, 2), 1.13)
        self.assertEqual(round(mean_p, 2), 1.13)


    ################
    # MC Diffusion #
    ################
    def test_diffusion_mc_model(self):
        # self.skipTest("Temporary")

        # Check cosine model
        model = pa.CosineModel("output/diff_mc_cyl_s.h5", 6, 10)

        # Check if the initialized profiles are corret
        self.assertTrue(np.array_equal(np.round(model._diff_bin,3), np.array([-3.702] * model._bin_num)))
        self.assertTrue(np.array_equal(model._df_bin, np.array([0] * model._bin_num)))

        # Check step model
        model = pa.StepModel("output/diff_mc_cyl_s.h5", 6, 10)

        # Check if the initialized profiles are corret
        self.assertTrue(np.array_equal(np.round(model._diff_bin,3), np.array([-3.702] * model._bin_num)))
        self.assertEqual(np.array_equal(model._df_bin, np.array([0] * model._bin_num)), True)
    #
    def test_diffusion_mc_mc(self):
        # self.skipTest("Temporary")

        # Set Cosine Model for diffusion and energy profile
        model = pa.CosineModel("output/diff_mc_cyl_s.h5", 6, 10)

        # Set the variable because this happen in the do_mc_cycles function -> not necessary to call to check the likelihood and Check if the initalize likelihood is correct
        pa.MC._len_step = 1
        self.assertEqual(round(pa.MC()._log_likelihood_z(model),2),-149411.12)
        # Set the variable because this happen in the do_mc_cycles function -> not necessary to call to check the likelihood and Check if the initalize likelihood is correct
        pa.MC._len_step = 2
        self.assertEqual(round(pa.MC()._log_likelihood_z(model),2),-169616.56)
        # Set the variable because this happen in the do_mc_cycles function -> not necessary to call to check the likelihood and Check if the initalize likelihood is correct
        pa.MC._len_step = 10
        self.assertEqual(round(pa.MC()._log_likelihood_z(model),2), -234093.97)

        # Set len_step for MC run test
        model._len_step = [10,20,40]

        #### Test Single ####
        # Do the MC alogirthm
        pa.MC().run(model,"output/diff_test_mc.yml", nmc_eq=1000, nmc=2000, is_print=False, is_parallel=False)

        # Plot diffusion coefficient over inverse lagtime
        diff = pa.diffusion.mc_fit("output/diff_test_mc.yml")
        plt.savefig("output/mc_fit.pdf", format="pdf", dpi=1000)

        # Plot pore diffusion coefficient over inverse lagtime
        plt.figure()
        diff_pore = pa.diffusion.mc_fit("output/diff_test_mc.yml", section="pore")
        plt.savefig("output/mc_fit_pore.pdf", format="pdf", dpi=1000)

        # Check if diffusion coefficient is in the range
        self.assertEqual(abs(diff[0] - (1.6) ) < 0.3, True)
        self.assertEqual(abs(diff_pore[0] - (1.2) ) < 0.3, True)

        #### Test Parallel ####
        # Do the MC alogirthm
        pa.MC().run(model,"output/diff_test_mc.h5", nmc_eq=8000, nmc=2000, is_print=False, is_parallel=True)

        # Plot diffusion coefficient over inverse lagtime
        plt.figure()
        diff = pa.diffusion.mc_fit("output/diff_test_mc.h5")
        plt.savefig("output/mc_fit.pdf", format="pdf", dpi=1000)

        # Plot pore diffusion coefficient over inverse lagtime
        plt.figure()
        diff_pore = pa.diffusion.mc_fit("output/diff_test_mc.h5", section="pore")
        plt.savefig("output/mc_fit_pore.pdf", format="pdf", dpi=1000)

        # Check if diffusion coefficient is in the range
        self.assertEqual(abs(diff[0]-(1.6) ) < 0.3, True)
        self.assertEqual(abs(diff_pore[0]-(1.2) ) < 0.3, True)

        # Test MC output
        # Set Step Model for diffusion and energy profile
        model = pa.StepModel("output/diff_mc_cyl_s.h5", 6, 10, is_print=True)

        # Set Cosine Model for diffusion and energy profile
        model = pa.CosineModel("output/diff_mc_cyl_s.h5", 6, 10, is_print=True)

        ## Set the MC class and options
        model._len_step = [10]

        # Do the MC alogirthm
        pa.MC().run(model,"output/diff_test_mc_spline.h5", nmc_eq=100, nmc=2000, is_print=True, is_parallel=False)

    def test_diffusion_mc_box(self):
        # self.skipTest("Temporary")

        # Set Cosine Model for diffusion and energy profile
        model = pa.CosineModel("output/diff_mc_box.h5", 6, 10)

        # Set the MC class and options
        model._len_step = [10, 20, 30, 40, 50]

        # Do the MC alogirthm
        pa.MC().run(model,"output/diff_test_mc_box.h5", nmc_eq=10000, nmc=3000, is_print=False, is_parallel=False)

        # Plot diffusion coefficient over inverse lagtime
        plt.figure()
        diff = pa.diffusion.mc_fit("output/diff_test_mc_box.h5")
        plt.savefig("output/diffusion_fit_box.pdf", format="pdf", dpi=1000)

        # Check if diffusion coefficient is in the range
        self.assertEqual(abs(diff[0] - (10.5) ) < 0.7, True)


    #####################
    # Transition Matrix #
    #####################
    def test_parallel_sample(self):
        # self.skipTest("Temporary")

        # Load Transition matrix for single
        data = pa.utils.load("output/diff_mc_cyl_s.h5")
        trans_s = data["data"]

        data = pa.utils.load("output/diff_mc_cyl_p.h5")
        trans_p = data["data"]

        data = pa.utils.load("data/check_output_sample.h5")
        trans_check = data["data"]

        # Test if transition matrix of single and parallel calculation is the same
        list = []
        list2 = []

        for i in [1,2,5,10,20,30,40]:
            list.append(np.array_equal(trans_s[i], trans_p[i]))
            list2.append(np.array_equal(trans_check[i], trans_p[i]))


    ###############
    # Free Energy #
    ###############
    def test_freeenergy_mc(self):
        # self.skipTest("Temporary")

        # Plot free energy profile over box length
        plt.figure()
        pa.freeenergy.mc_profile("data/check_output.h5")
        plt.savefig("output/energy_profile.pdf", format="pdf", dpi=1000)


    ##########
    # Tables #
    ##########
    def test_tables(self):
        # self.skipTest("Temporary")

        # Check tables
        pa.tables.mc_model("data/check_output.h5", print_con=False)
        pa.tables.mc_model("data/box_output.h5", print_con=False)
        pa.tables.mc_inputs("data/check_output.h5", print_con=False)
        pa.tables.mc_statistics("data/check_output.h5", print_con=False)
        pa.tables.mc_lag_time("data/check_output.h5", print_con=False)
        pa.tables.mc_results("data/check_output.h5", print_con=False)

        pa.tables.mc_model("data/check_output.h5", print_con=True)
        pa.tables.mc_inputs("data/check_output.h5", print_con=True)
        pa.tables.mc_statistics("data/check_output.h5", print_con=True)
        pa.tables.mc_lag_time("data/check_output.h5", print_con=True)
        pa.tables.mc_results("data/check_output.h5", print_con=True)
        pa.tables.mc_results("data/box_output.h5", print_con=True)
        pa.tables.mc_results("data/check_output.h5", print_con=True, sections={"test": [0,5]})
        pa.tables.mc_results("data/box_output.h5", print_con=True, sections={"test": [0,5]})

        print()
        self.assertIsNone(pa.tables.mc_results("data/box_output.h5", sections={"DOTA": []}))

    def test_diffusion_output(self):
        # self.skipTest("Temporary")

        # Check output which is not coveraged by the entire MC test
        # Check diffusion profile function
        pa.diffusion.mc_profile("data/check_output.h5", len_step=[10,20,40], infty_profile = False)
        pa.diffusion.mc_profile("data/check_output.h5", len_step=[10,20,40], section = "pore", infty_profile = True)
        pa.diffusion.mc_profile("data/check_output.h5", len_step=[10,20,40], section = "reservoir", infty_profile = True)
        pa.diffusion.mc_profile("data/check_output.h5", len_step=[10,20,40], section = "test", infty_profile = True)
        pa.diffusion.mc_profile("data/check_output.h5", len_step=[10,20,40], section = [0,1,2], infty_profile = True)
        pa.diffusion.mc_profile("data/check_output.h5", section = [1,10], infty_profile = True)

        # Check diffusion fitting function
        pa.diffusion.mc_fit("data/check_output.h5", section = "pore")
        pa.diffusion.mc_fit("data/check_output.h5", section = "reservoir")
        pa.diffusion.mc_fit("data/check_output.h5", section = "test")
        pa.diffusion.mc_fit("data/check_output.h5", section = [0,1,2])
        pa.diffusion.mc_fit("data/check_output.h5", section=[0,10])

        # Check transition matrix heatmap
        pa.diffusion.mc_trans_mat("data/check_output.h5",10)
        pa.diffusion.mc_trans_mat("data/check_output_sample.h5",10)

        # Check if box not pore system
        pa.diffusion.mc_fit("data/box_output.h5", section = "pore")
        pa.diffusion.mc_fit("data/box_output.h5", section = "reservoir")
        pa.diffusion.mc_profile("data/box_output.h5", section = "pore")
        pa.diffusion.mc_profile("data/box_output.h5", section = "reservoir")



if __name__ == '__main__':
    unittest.main(verbosity=2)



################################################################################
# Tables                                                                       #
#                                                                              #
"""Print analysis information in table form."""
################################################################################


import numpy as np
import pandas as pd
import poreana.utils as utils
import poreana.diffusion as diffusion

###############################
# MC - Diffusion, Free Energy #
###############################
def mc_statistics(link, print_con=False):
    """This function prints the statistic of the MC Run for every lag time.

    Parameters
    ----------
    link : string
        Link to the diffusion hdf5, obj or yml data file generated by the MC Algorithm function
        :func:`poreana.mc.MC.run`
    print_con: bool, optional
        True for printing in console

    Returns
    -------
    df_results: obj
        Data frame of the MC Alogrithm statistics
    """
    # Load Results from the output object file
    data = utils.load(link)

    # Load results
    results = data["output"]

    # Load model inputs
    model = data["model"]
    len_step = model["len_step"]
    inp = data["inp"]
    nmc_eq = inp["MC steps eq"]
    nmc = inp["MC steps"]


    nacc_df_mean = results["nacc_df"]
    nacc_diff_mean = results["nacc_diff"]
    list_diff_fluc = results["fluc_diff"]
    list_df_fluc = results["fluc_df"]

    # Table for MC Statistics
    data = [[str("%.4e" % list_df_fluc[i]) for i in len_step],[str("%.4e" % list_diff_fluc[i]) for i in len_step],[str("%.0f" % nacc_df_mean[i]) for i in len_step],[str("%.0f" % nacc_diff_mean[i]) for i in len_step],[str("%.2f" % (nacc_df_mean[i]*100/(nmc+nmc_eq))) for i in len_step],[str("%.2f" % (nacc_diff_mean[i]*100/(nmc+nmc_eq))) for i in len_step]]

    df_results = pd.DataFrame(data,index=list(['fluctuation df','fluctuation diff','acc df steps','acc diff steps','acc df steps (%)','acc diff steps (%)']),columns=list(len_step))

    # If the table has to print in console
    if print_con:
        print('\nStatistics of the MC Algorithm')
        print(df_results)

    # Set styler for pandas table in jupyter
    df_results = pd.DataFrame(df_results.rename_axis('Step Length', axis=1))
    styler = df_results.style.set_caption('Statistics of the MC Algorithm')
    df_results = styler.set_properties(**{'text-align': 'right'})
    df_results = df_results.set_table_styles([dict(selector='th', props=[('text-align', 'left')])])

    return df_results


def mc_lag_time(link, print_con=False):
    """This function prints the final coefficients of the profile for
    every lag time profile.

    Parameters
    ----------
    link : string
        Link to the diffusion hdf5, obj or yml data file generated by the MC Algorithm function
        :func:`poreana.mc.MC.run`
    print_con: bool, optional
        True for printing in console

    Returns
    -------
    df_results: obj
        Data frame of profile coefficients
    """
    # Load Results from the output object file
    data = utils.load(link)

    # Load results
    results = data["output"]
    diff_coeff = results["diff_coeff"]
    df_coeff = results["df_coeff"]

    # Load model inputs
    model = data["model"]
    len_step = model["len_step"]
    nD = model["nD"]
    nF = model["nF"]

    # Initialize data dictionary for the diffusion profile coefficients
    data = {}

    # Save diffusion profile coefficients on data dictionary
    for i in len_step:
        data[i] = [str("%.4e" % diff_coeff[i][j]) for j in range(int(nD))]

    # Pandas table
    diff_coeff = pd.DataFrame(data, index=list(np.arange(1, int(nD)+1)), columns=list(len_step))
    diff_coeff = pd.DataFrame(diff_coeff.rename_axis('Step Length', axis=1))

    # If the table has to print in console
    if print_con:
        print('\nDiffusion coefficients')
        print(diff_coeff)

    # Set styler for pandas table in jupyter
    styler = diff_coeff.style.set_caption('Diffusion coefficients')
    diff_coeff = styler.set_properties(**{'text-align': 'right'})
    diff_coeff = diff_coeff.set_table_styles([dict(selector='th', props=[('text-align', 'left')])])

    # Initialize data dictionary for the diffusion profile coefficients
    data = {}

    # Save free energy profile coefficients on data dictionary
    for i in len_step:
        data[i] = [str("%.4e" % df_coeff[i][j]) for j in range(int(nF))]

    # Pandas table
    df_coeff = pd.DataFrame(data, index=list(np.arange(1, int(nF)+1)), columns=list(len_step))
    df_coeff = pd.DataFrame(df_coeff.rename_axis('Step Length', axis=1))

    # If the table has to print in console and not in a jupyter notebook
    if print_con:
        print('\nFree energy coefficients')
        print(df_coeff)

    # Set styler for pandas table in jupyter
    styler = df_coeff.style.set_caption('Free energy coefficients')
    df_coeff = styler.set_properties(**{'text-align': 'right'})
    df_coeff = df_coeff.set_table_styles([dict(selector='th', props=[('text-align', 'left')])])

    return diff_coeff, df_coeff


def mc_model(link, print_con=False):
    """This function prints the model inputs of the calculation.

    Parameters
    ----------
    link : string
        Link to the diffusion hdf5, obj or yml data file generated by the MC Algorithm function
        :func:`poreana.mc.MC.run`
    print_con: bool, optional
        True for printing in console

    Returns
    -------
    df_model: obj
        Data frame of the model inputs
    """
    # Load Results from the output object file
    data = utils.load(link)

    # Load model inputs
    model = data["model"]
    bin_number = model["bin number"]
    len_step = model["len_step"]
    len_frame = model["len_frame"]
    frame_num = model["num_frame"]
    nD = model["nD"]
    nF = model["nF"]
    nDrad = model["nDrad"]
    d = model["guess"]
    model_string = model["model"]
    pbc = model["pbc"]
    direction = int(model["direction"])

    # String for pbc
    pbc = pbc == 1

    # String for system
    system = "pore" if "pore" in data else "box"

    # Len step string
    len_step_string = ', '.join(str(step) for step in len_step)

    # Dictionary for model inputs
    data = [str("%.i" % bin_number), len_step_string, str("%.2e" % (len_frame * 10**(-12))), str("%.i" % frame_num), str("%.i" % nD), str("%.i" % nF), str("%.i" % nDrad), model_string, str("%.2e" % (d * 10**(-6))), system, pbc, str("%.i" % direction)]
    df_model = pd.DataFrame(data, index=list(['Bin number', 'step length', 'frame length (s)', 'frame number', 'nD', 'nF', 'nDrad', 'model', 'guess diffusion (m2/s-1)', 'system',"pbc", "direction"]), columns=list(['Input']))

    # If the table has to print in console and not in a jupyter notebook
    if print_con:
        print('\nModel Inputs')
        print(df_model)

    # Set styler for pandas table in jupyter
    #styler = df_model.style.set_caption('Model Inputs')
    #df_model = styler.set_properties(**{'text-align': 'right'})
    #df_model = df_model.set_table_styles([dict(selector='th', props=[('text-align', 'left')])])

    return df_model


def mc_inputs(link, print_con=False):
    """This function prints the MC Algorithm inputs of the calculation.

    Parameters
    ----------
    link : string
        Link to the diffusion hdf5, obj or yml data file generated by the MC Algorithm function
        :func:`poreana.mc.MC.run`
    print_con : bool, optional
        True for printing in console

    Returns
    -------
    df_mc : obj
        Data frame of the MC Alogrithm inputs
    """
    # Load Results from the output object file
    data = utils.load(link)


    # Read MC inputs
    inp = data["inp"]
    nmc_eq = inp["MC steps eq"]
    nmc = inp["MC steps"]
    num_mc_update = inp["step width update"]
    print_freq = inp["print freq"]
    temp = inp["temperature"]

    # Table for MC Inputs
    data = [str("%.i" % nmc_eq), str("%.i" % nmc), temp, str("%.i" % num_mc_update), str("%.i" % print_freq)]
    df_mc = pd.DataFrame(data, index=list(['MC steps (Equilibrium)', 'MC steps (Production)','temperature (MC)', 'movewidth update frequency', 'print frequency']), columns=list(['Input']))

    # If the table has to print in console and not in a jupyter notebook
    if print_con:
        print('\nMC Inputs')
        print(df_mc)

    # Set style for the pandas table
    #styler = df_mc.style.set_caption('MC Inputs')
    #df_mc = styler.set_properties(**{'text-align': 'right'})
    #df_mc = df_mc.set_table_styles([dict(selector='th', props=[('text-align', 'left')])])

    return df_mc

def mc_results(link, print_con=False, sections={"box": [], "pore": [], "res": []}, len_step = []):
    """This function prints the MC results of the diffusion calculation.
    Different areas are defined in the **sections** dectionary. Standard areas
    are calculated by leaving the value list empty

    * **box** - For the entire system
    * **pore** - Pore section (only for pore systems)
    * **res** - Resercoir section (only for pore systems)

    Parameters
    ----------
    link : string
        Link to the diffusion hdf5, obj or yml data file generated by the MC Algorithm
        function :func:`poreana.mc.MC.run`
    print_con : bool, optional
        True for printing in console
    sections : dictionary, optinal
        Sections for diffusion calculation

    Returns
    -------
    df_mc_results : DataFrame
        Data frame of the MC Alogrithm inputs
    """
    # Load data
    data = utils.load(link)
    model = data["model"]

    # Process input
    len_step = list(model["len_step"]) if not len_step else len_step

    # Extract system data
    box_z = round(data["pore"]["box"][2], 2) if "pore" in data else round(data["box"]["length"][2], 2)
    res = round(data["pore"]["res"], 2) if "pore" in data else 0

    # Run through sections
    diff_dict = {}
    for key, section in sections.items():
        # Process special cases
        if not section:
            if key=="box":
                area = [0, box_z]
            elif key=="pore":
                area = [res, round(box_z-res, 2)] if "pore" in data else []
            elif key=="res":
                area = [0, res] if "pore" in data else []
            else:
                print("The area of \""+key+"\" needs to be specified")
                return
        else:
            area = section

        # Calculate diffusion in area
        if area:
            diff = diffusion.mc_fit(link, section=area, is_print=False, is_plot=False, len_step=len_step)
            diff_dict[key] = {"Area (nm)": area, "Diffusion (10^-9 m^2s^-1)": diff[0], "Residual (10^-9 m2s^-1)": diff[3]}

    # Create pandas table
    df_mc_results = pd.DataFrame.from_dict(diff_dict, orient="index")

    # If the table has to print in console and not in a jupyter notebook
    if print_con:
        print('\nMC Results')
        print(df_mc_results)

    return df_mc_results



################################################################################
# Free Energy                                                                  #
#                                                                              #
"""Analyse free energy in a pore."""
################################################################################


import seaborn as sns
import poreana.utils as utils
import matplotlib.pyplot as plt


####################
# Free Energy - MC #
####################
def mc_profile(link, len_step=[], is_plot=True, kwargs={}):
    """This function plots the free energy profile over the box for the
    calculated lag times. In contrast to the diffusion profile the diffusion
    profile has not a dependency on the lag time. If the free energy profiles
    are not close to equal the calculation is incorrect.

    Parameters
    ----------
    link : string
        Link to the diffusion hdf5, obj or yml data file generated by the :func:`poreana.mc.MC.do_mc_cycles`
    len_step: integer list, optional
        List of the different step length, if it is [] all free energy profiles
        depending on the lag time are shown
    is_plot : bool, optional
        Show free energy profile
    kwargs: dict, optional
        Dictionary with plotting parameters

    Returns
    -------
    df_bin: dictionary
        free energy profile for every calculated lag time
    bins : list
        bins over the box length
    """
    # Load Results from the output object file
    data = utils.load(link)

    # Load results
    results = data["output"]
    df_bin = results["df_profile"]

    # Load model inputs
    model = data["model"]
    dt = model["len_frame"]
    bins = model["bins"]

    # If no specific step length is chosen take the step length from the object file
    if not len_step:
        len_step = model["len_step"]

    # Set legend
    legend = ["$\\Delta t_{\\alpha}$ = " + str(len_step[i] * dt) + " ps" for i in range(len(len_step))]


    # Plot the free energy profiles
    if is_plot:
        for i in len_step:
            sns.lineplot(x=bins, y=(df_bin[i]), **kwargs)

        # Plot options
        plt.xlabel("Box length (nm)")
        plt.ylabel("Free energy (-)")
        plt.legend(legend)
        plt.xlim([0,max(bins)])

    return df_bin, bins



import numpy as np
import pandas as pd
import poreana.utils as utils


class Model:
    """This class sets the general parameters which are used to initialize
    a model.

    Parameters
    ----------
    link : string
        Data link to the hdf5, obj or yml data file from :func:`poreana.sample.Sample.init_diffusion_mc`

    """

    def __init__(self, link):
        # Load hdf5 data file
        sample = utils.load(link)

        inp = sample["inp"]
        # Read the inputs
        self._bin_num = int(inp["bin_num"])               # number of bins z-direction
        self._frame_num = int(inp["num_frame"])             # number of frames
        self._len_step = inp["len_step"]                 # step length
        self._dt = float(inp["len_frame"]) * 10**12             # frame length [ps]
        self._bins = inp["bins"]                      # bins [nm]
        self._bin_width = self._bins[1] - self._bins[0]  # bin width [nm]
        self._direction = int(inp["direction"])
        self._trans_mat= sample["data"]                # transition matrix
        self._pbc = inp["is_pbc"]                    # pbc or nopbc
        self._type = sample["type"]

        self._sys_props = {}
        if "pore" in sample:
            self._sys_props["type"] = sample["pore"]["type"]
            self._sys_props["res"] = float(sample["pore"]["res"])
            self._sys_props["focal"] = sample["pore"]["focal"]
            self._sys_props["box"] = sample["pore"]["box"]
            self._sys_props["diam"] = float(sample["pore"]["diam"])
            self._system = "pore"
        if "box" in sample:
            self._system = "box"
            self._sys_props["length"] = sample["box"]["length"]


        # Initialize units of diffusion and free energy unit
        self._df_unit = 1.                                 # in kBT
        self._diff_unit = np.log(self._bin_width**2 / 1.)  # in m^2/s

    def _init_profiles(self):
        """This function initializes the normal diffusion, radial diffusion and
        free energy profile over the bins.
        """
        # Initialize the diffusion and free energy profile
        self._df_bin = np.float64(np.zeros(self._bin_num))    # in kBT
        self._diff_bin = np.float64(np.zeros(self._bin_num))  # in dz**2/dt

        # Initialize the diffusion profile
        self._diff_bin += (np.log(self._d0) - self._diff_unit)

    def _calc_profile(self, coeff, basis):
        """This function calculates the diffusion and free energy profile over
        the bins. It is used to initialize the system at the beginning of the
        calculation/MC run. Additionally, it is needed to update the profiles
        in the Monte Carlo part after the adjustment of a profile coefficient.

        The profile is determining with the basis and the coefficients for the
        diffusion with

        .. math::

            \\ln D_i = a_0 + \\sum_{k} a_{k} \\cdot \\mathrm{basis}_{\\text{F}},

        and for the free energy with

        .. math::

            F_i = \\sum_{k} a_{k} \\cdot \\mathrm{basis}_{\\text{D}}.

        The diffusion is calculated between to bins on the bin border and the free energy
        in the center of a bin.
        The basis is received by the create basis function for the selected model:

        **CosineModel**
         * :func:`CosineModel._create_basis_center`
         * :func:`CosineModel._create_basis_border`

        **StepModel**
         * :func:`StepModel._create_basis_center`
         * :func:`StepModel._create_basis_border`

        Parameters
        ----------
        coeff : list
            list of coefficients
        basis : list
            list of the basis part of model
        """

        # Calculate a matrix with the bin_num x ncos
        # Columns contains the n-summand in every bin
        # Line contains the summand of the series
        a = coeff * basis

        return np.sum(a, axis=1)


class CosineModel(Model):
    """This class sets the Cosine Model to calculate the free energy profile
    and the diffusion profile. The profiles have the typical cosine oscillation.
    These profiles over the bins are expressed by the following Fourier
    series. The diffusion profile is calculated between bin i and i+1  over
    the bin border with

    .. math::

        \\ln \\left(D_{i+ \\frac{1}{2}}\\right)=a_{0}+\\sum_{k=1}^{n_b} a_{k} \\cdot \\cos \\left(\\frac{2\\pi ki}{n}\\right).

    Similarly, the free energy Fourier series can be written:

    .. math::

        F_{i} = a_{0}+\\sum_{k=1}^{n_b} a_{k} \\cdot \\cos \\left(\\frac{2\\pi ki}{n} \\right),

    with the number of bins :math:`n`, the index of a bin :math:`i`,
    the coefficients :math:`a_{k}` of the Fourier series and :math:`k` as the
    number of coefficents. The free energy is calculated in the bin center.

    For the free energy the coefficient it is assumed that :math:`a_{0} = 0`.

    Parameters
    ----------
    data_link : string
        Data link to the pickle data from :func:`poreana.sample.Sample.init_diffusion_mc`
    n_diff : integer, optional
        Number of the Fourier coefficients for the diffusion profile
    n_df : integer, optional
        Number of the Fourier coefficients for the free energy profile
    n_diff_radial : integer, optional
        Number of the Fourier coefficients for the radial diffusion profile
    d0 : double, optional
        Initial guess of diffusion coefficient :math:`\\left( 10^{-9} \\frac{m^2}{s}\\right)`
    is_print : bool, optional
        True to print output
    """

    def __init__(self, data_link, n_diff=6, n_df=10, n_diff_radial=6, d0=1, is_print=False):

        # Inherit the variables from Model class
        super(CosineModel, self).__init__(data_link)

        # Set the model type
        self._model = "CosineModel"

        self._n_df = n_df
        self._n_diff = n_diff
        self._n_diff_radial = n_diff_radial
        self._print_output = is_print
        self._d0 = d0 * (10**9)/(10**12)                # guess init profile [A^2/ps]

        self._init_model()     # Initial model
        self._init_profiles()  # Initial Profiles
        self._cosine_model()   # Set basis of Fourier series

    def _init_model(self):
        """This function initializes the coefficient list for the Fourier series
        and the profile list for the free energy and diffusion profile.
        It is used to reinitialize these lists at the beginning of a MC run.
        """

        # Initialize the diffusion and free energy coefficient
        self._df_coeff = np.float64(np.zeros(self._n_df))      # in dz**2/dt
        self._diff_coeff = np.float64(np.zeros(self._n_diff))  # in dz**2/dt

        # Set start diffusion profile
        self._diff_coeff = np.zeros((self._n_diff), float)

        # Initialize diffusion profile with the guess value [A^2/ps]
        self._diff_coeff[0] += (np.log(self._d0) - self._diff_unit)

    def _cosine_model(self):
        """This function sets a Fourier Cosine Series Model for the MC Diffusion
        Calculation and determines the initialize profiles.
        """

        # Create basis (for the free energy)
        self._create_basis_center()

        # Create basis (for the free energy)
        self._create_basis_border()

        # Update diffusion profile
        self._diff_bin = self._calc_profile(self._diff_coeff, self._diff_basis)

        # Update free energy profile
        self._df_bin = self._calc_profile(self._df_coeff, self._df_basis)

        # Print for console
        if self._print_output == True:
            print("--------------------------------------------------------------------------------")
            print("-------------------------Initalize Cosine Model---------------------------------")
            print("--------------------------------------------------------------------------------\n")
            print("Model Inputs")

            # Set data list for panda table
            len_step_string = ', '.join(str(step) for step in self._len_step)
            data = [str("%.f" % self._bin_num),  len_step_string, str("%.2e" % (self._dt * 10**(-12))), str("%.f" % self._n_diff),
                    str("%.f" % self._n_df), self._model, self._pbc, str("%.2e" % (self._d0 * (10**(-18))/(10**(-12)))), self._system]

            # Set pandas table
            df_model = pd.DataFrame(data, index=list(
                ['Bin number', 'step length', 'frame length', 'nD', 'nF', 'model', 'pbc', 'guess diffusion (m2/s-1)', 'system']), columns=list(['Input']))

            # Print panda table with model inputs
            print(df_model)

    def _create_basis_center(self):
        """This function creates the basis part of the Fourier series for the
        free energy and the radial diffusion profile.
        For a bin the basis is calculated with

        .. math::

            \\mathrm{basis} = \\cos \\left(\\frac{2\\pi k(i+0.5)}{n}\\right)

        hereby :math:`k` is the number of coefficients, :math:`i` is the bin
        index and :math:`n` is the number of the bins.
        """

        # Allocate a vector with bin_num entries
        x = np.arange(self._bin_num)

        # Calculate basis for Fourier cosine series
        basis_df = [np.cos(2 * k * np.pi * (x + 0.5) / self._bin_num) / (k + 1) for k in range(self._n_df)]                      # basis for the free energy profile
        basis_diff_radial = [np.cos(2 * k * np.pi * (x + 0.5) / self._bin_num) / (k + 1) for k in range(self._n_diff_radial)]    # basis for the radial energy profile

        # Transpose basis (is now a bin_num x ncos Matrix)
        self._df_basis = np.array(basis_df).transpose()
        self._diff_radial_basis = np.array(basis_diff_radial).transpose()

    def _create_basis_border(self):
        """This function creates the basis part in every bin of the Fourier
        series for the diffusion :math:`\\ln \\ (D)`.
        At the bin border the basis is calculated with

        .. math::

            \\mathrm{basis} = \\cos \\left(\\frac{2\\pi ki}{n}\\right)

        hereby :math:`k` is the number of coefficients, :math:`i` is the bin
        index and :math:`n` is the number of the bins.
        """

        # Allocate a vector with bin_num entries
        x = np.arange(self._bin_num)

        # Calculate basis for Fourier cosine series
        basis = [np.cos(2 * k * np.pi * (x + 1.) / self._bin_num) / (k + 1) for k in range(self._n_diff)]

        # Transpose basis (is now a bin_num x ncos Matrix)
        self._diff_basis = np.array(basis).transpose()


class StepModel(Model):
    """This class sets the Step Model to calculate the free energy profile and
    the diffusion profile. This model based on a spline calculation.
    In contrast to the Cosine Model the determined profile has not the typical
    oscillation and receives a profile which is better interpretable.
    The profile will be determined with the following equations

    .. math::

        \\ln \\left(D_{i+\\frac{1}{2}}\\right) = a_{k} \\cdot \\mathrm{basis}_{\\mathrm{diff}}

    .. math::

        F_i = a_{k} \\cdot \\mathrm{basis}_{\\mathrm{df}}

    with :math:`a_k` as the coefficients and the :math:`\\mathrm{basis}`. The
    basis of the modell is calculated with

    **StepModel**

    * :func:`StepModel._create_basis_center`
    * :func:`StepModel._create_basis_border`

    Parameters
    ----------
    data_link : string
        Data link to the pickle data from :func:`poreana.sample.Sample.init_diffusion_mc`
    n_diff : integer, optional
        Number of the Fourier coefficients for the diffusion profile
    n_df : integer, optional
        Number of the Fourier coefficients for the free energy profile
    n_diff_radial : integer, optional
        Number of the Fourier coefficients for the radial diffusion profile
    d0 : double, optional
        Initial guess of diffusion coefficient :math:`\\left( 10^{-9} \\frac{m^2}{s}\\right)`
    is_print : bool, optional
        True to print output
    """

    def __init__(self, data_link, n_diff=6, n_df=10, n_diff_radial=6, d0=1, is_print=False):

        # Inherit the variables from Model class
        super(StepModel, self).__init__(data_link)

        self._model = "Step Model"

        self._n_diff = n_diff
        self._n_df = n_df
        self._n_diff_radial = n_diff_radial
        self._print_output = is_print
        self._d0 = d0 * (10**9)/(10**12)

        self._init_model()     # Initial model
        self._init_profiles()  # Initial Profiles
        self._step_model()     # Set basis of Step Model

    def _init_model(self):
        """This function initializes the coefficient list for the Step Model and
        the profile list for the free energy and diffusion profile. It is used
        to reinitializes these lists for every lag time calculation.
        """

        # Initialize the diffusion and free energy coefficient
        self._df_coeff = np.float64(np.zeros(self._n_df))
        self._diff_coeff = np.float64(np.zeros(self._n_diff))

        # Calculate dz
        dx_df = self._bin_num / 2. / self._n_df
        dx_diff = self._bin_num / 2. / self._n_diff

        self._df_x0 = np.arange(0, self._n_df * dx_df, dx_df)
        self._diff_x0 = np.arange(0, self._n_diff * dx_diff, dx_diff)

        # Initialize diffusion profile with the guess value [A^2/ps]
        self._diff_coeff[0] += (np.log(self._d0) - self._diff_unit)

    def _step_model(self):
        """This function set a Step Model for the MC Diffusion Calculation and
        determine the initialize profiles.
        """

        # create basis (for the free energy)
        self._create_basis_center()

        # create basis (for the free energy)
        self._create_basis_border()

        # Update diffusion profile
        self._diff_bin = self._calc_profile(self._diff_coeff, self._diff_basis)

        # Update free energy profile
        self._df_bin = self._calc_profile(self._df_coeff, self._df_basis)

        # Print for console
        if self._print_output == True:
            print("\n")
            print("--------------------------------------------------------------------------------")
            print("-------------------------Initalize Step Model-----------------------------------")
            print("--------------------------------------------------------------------------------\n")
            print("Model Inputs")

            # Set data list for panda table
            len_step_string = ', '.join(str(step) for step in self._len_step)
            data = [str("%.f" % self._bin_num),  len_step_string, str("%.2e" % (self._dt * 10**(-12))), str("%.f" % self._n_diff),
                    str("%.f" % self._n_df), self._model, self._pbc, str("%.2e" % (self._d0 * (10**(-18))/(10**(-12)))), self._system]

            # Set pandas table
            df_model = pd.DataFrame(data, index=list(
                ['Bin number', 'step length', 'frame length', 'nD', 'nF', 'model', 'pbc', 'guess diffusion (m2/s-1)','system']), columns=list(['Input']))

            # Print panda table with model inputs
            print(df_model)
            print("\n")

    def _create_basis_center(self):
        """This function creates the basis part in every bin of the Step model
        for the free energy and the radial diffusion profile. The following
        explanation is for the free energy profile. For the radial diffusion
        profile the number of free energy coefficients :math:`n_{\\mathrm{df}}` has to
        exchange with :math:`n_{\\mathrm{diff\_radial}}`. The Dimension of the basis matrix
        is :math:`n_{\\mathrm{bin}} \\times n_{\\mathrm{df}}`. For a bin the
        basis is calculated with

        .. math::

            \\mathrm{basis} = \\begin{cases}
                            1 & (\\mathrm{bin}+0.5)\\geq \\Delta x\ \\& \ (\\mathrm{bin}+0.5)\\leq n_{\\mathrm{bin}}-\\Delta x \\\\
                            0 & \\mathrm{else}                                 \\
                    \\end{cases}

        hereby is :math:`\\mathrm{bin} = [0,...,n_{bin}]` with
        :math:`n_{\\mathrm{bin}}` as the number of the bins. The variable
        :math:`\\Delta x` is define by

        .. math::

            \\Delta x = \\left [ 0,i \\cdot \\frac{n_{\\mathrm{bin}}}{n_{\\mathrm{df}} \\cdot 2},\\frac{n_{\\mathrm{bin}}}{2} \\right ]

        with :math:`i = [1,...,n_{\\mathrm{df}}-1]`.
        """
        # Calculated the basis in the center of a bin
        x = np.arange(self._bin_num)+0.5
        basis = [np.where((x >= i) & (x <= self._bin_num-i), 1., 0.) for i in self._df_x0]

        # Transpose basis (is now a bin_num x ncos Matrix)
        self._df_basis = np.array(basis).transpose()

    def _create_basis_border(self):
        """This function creates the basis part in every bin of the Step model
        for the diffusion profile. The dimension of the basis matrix is
        :math:`n_{bin} \\times n_{diff}`. At the bin border the basis is
        calculated with

        .. math::

            \\mathrm{basis} = \\begin{cases}
                            1 & (\\mathrm{bin}+1)\\geq \\Delta x\ \\& \ (\\mathrm{bin}+1)\\leq n_{\\mathrm{bin}}-\\Delta x \\\\
                            0 &  \\mathrm{else}                                 \\
                    \\end{cases}

        hereby is :math:`bin = [0,...,n_{\\mathrm{bin}}]` with
        :math:`n_{\\mathrm{bin}}` as the number of the bins. The variable
        :math:`\\Delta x` is define by

        .. math::

            \\Delta x = \\left [ 0,i \\cdot \\frac{n_{\\mathrm{bin}}}{n_{\\mathrm{diff}} \\cdot 2},\\frac{n_{\\mathrm{bin}}}{2} \\right ]

        with :math:`i = [1,...,n_{\\mathrm{diff}}-1]`.
        """
        # Calculated the basis in the border of a bin
        x = np.arange(self._bin_num)+1.
        basis = [np.where((x >= i) & (x <= self._bin_num-i), 1., 0.) for i in self._diff_x0]

        # Transpose basis (is now a bin_num x ncos Matrix)
        self._diff_basis = np.array(basis).transpose()
